import os
import bisect
from pwact.utils.file_operation import search_files, link_file, copy_file, write_to_file
from pwact.utils.constant import INIT_BULK, DFT_STYLE, TEMP_STRUCTURE

from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.init_bulk_input import InitBulkParam

from pwact.data_format.configop import do_super_cell, do_scale, do_pertub

def  duplicate_scale(resource: Resource, input_param:InitBulkParam):
    #work dir
    init_configs = input_param.sys_config
    relax_dir = os.path.join(input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.relax)
    duplicate_scale_dir = os.path.join(input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.super_cell_scale) 
                 
    for init_config in init_configs:
        # print(init_config.super_cell)
        config_format = init_config.format#
        config_file = init_config.config_file
        init_config_name = "init_config_{}".format(init_config.config_index)
        if init_config.relax:
            config_file = os.path.join(relax_dir, init_config_name, DFT_STYLE.get_relaxed_config(resource.dft_style))
            config_format = DFT_STYLE.get_format_by_postfix(DFT_STYLE.get_relaxed_config(resource.dft_style))
        else:
            config_file = init_config.config_file
        super_cell_scale_dir = os.path.join(duplicate_scale_dir, init_config_name)
        super_cell_config = None
        if init_config.super_cell is not None:
            if not os.path.exists(super_cell_scale_dir):
                os.makedirs(super_cell_scale_dir)
            # super_content, lattic_index = super_cell(init_config.super_cell, config, super_cell_config)

            super_cell_config = os.path.join(super_cell_scale_dir, DFT_STYLE.get_super_cell_config(resource.dft_style))

            if not os.path.exists(super_cell_config):
                do_super_cell(config_file=config_file,
                    input_format=config_format,
                    supercell_matrix=init_config.super_cell, 
                    pbc=init_config.pbc, 
                    direct = True, 
                    sort = True,
                    save_format=DFT_STYLE.get_pwdata_format(resource.dft_style), #cp2k ->vasp/poscar
                    save_path=super_cell_scale_dir, 
                    save_name=DFT_STYLE.get_super_cell_config(resource.dft_style))
            config_format = DFT_STYLE.get_pwdata_format(resource.dft_style)
            
            
        if init_config.scale is not None:
            for s_index, scale in enumerate(init_config.scale):
                # save_scale_config = os.path.join(super_cell_scale_dir, "{}_{}".format(scale, DFT_STYLE.get_scale_config(init_config.dft_style)))
                if not os.path.exists(super_cell_scale_dir):
                    os.makedirs(super_cell_scale_dir)
                if super_cell_config is not None:
                    input_scale_config = super_cell_config # config from super cell
                else:
                    input_scale_config = config_file # config from relax or init config
                do_scale(config=input_scale_config,
                    input_format = config_format,
                    scale_factor=scale, 
                    direct=True,
                    sort=True, 
                    save_format=DFT_STYLE.get_pwdata_format(resource.dft_style),
                    save_path=super_cell_scale_dir, 
                    save_name="{}_{}".format(scale, DFT_STYLE.get_scale_config(resource.dft_style))
                    )

def do_post_duplicate_scale(resource: Resource, input_param:InitBulkParam):
    duplicate_scale_dir = os.path.join(input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.super_cell_scale) 
    real_duplicate_scale_dir = os.path.join(input_param.root_dir, INIT_BULK.super_cell_scale) 
    if os.path.exists(duplicate_scale_dir):
        link_file(duplicate_scale_dir, real_duplicate_scale_dir)
    tag = os.path.join(duplicate_scale_dir, INIT_BULK.tag_super_cell)
    if not os.path.exists(tag):
        write_to_file(tag, "super cell and scale done ", "w")

def duplicate_scale_done(input_param:InitBulkParam):
    duplicate_scale_dir = os.path.join(input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.super_cell_scale) 
    tag = os.path.join(duplicate_scale_dir, INIT_BULK.tag_super_cell)
    return True if os.path.exists(tag) else False

'''
description: 
param {str} work_dir the dir of atom.config file
param {int} pert_num
param {float} cell_pert_fraction
param {float} atom_pert_distance
return {*}
author: wuxingxing
'''
def do_pertub_work(resource: Resource, input_param:InitBulkParam):
    #work dir 
    # pert_num:int=50
    # cell_pert_fraction:float=0.03
    # atom_pert_distance:float=0.01
    init_configs = input_param.sys_config
    relax_dir = os.path.join(input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.relax)
    duplicate_scale_dir = os.path.join(input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.super_cell_scale) 
    pertub_dir = os.path.join(input_param.root_dir,TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.pertub)
    # real_pertub_dir = os.path.join(input_param.root_dir,INIT_BULK.pertub)
    # make pertub dirs
    for init_config in init_configs:
        if init_config.perturb is None:
            continue
        init_config_name = "init_config_{}".format(init_config.config_index)
        pert_config_list, config_type = get_config_files_with_order(
            super_cell_scale_dir=duplicate_scale_dir,
            relax_dir=relax_dir,
            init_config_dirname=init_config_name, 
            init_config_path=init_config.config_file, 
            pertub_dir=None,
            dft_style=resource.dft_style
            )
        print(pert_config_list)
        for index, config in enumerate(pert_config_list):
            if config_type == INIT_BULK.scale: 
                tmp_config_dir = os.path.basename(os.path.basename(config).replace(DFT_STYLE.get_postfix(resource.dft_style), "")) # 0.8_scale.poscar or 0.8_scale.config -> 0.8_scale
            else:
                tmp_config_dir = config_type
            work_dir = os.path.join(pertub_dir, init_config_name, tmp_config_dir)
            if not os.path.exists(work_dir):
                os.makedirs(work_dir)
            target_config = os.path.join(work_dir, os.path.basename(config))
            copy_file(config, target_config)
            
            do_pertub(config=target_config, 
                input_format=DFT_STYLE.get_format_by_postfix(os.path.basename(target_config)),
                pert_num=init_config.perturb, 
                cell_pert_fraction=init_config.cell_pert_fraction, 
                atom_pert_distance=init_config.atom_pert_distance, 
                direct=True,
                sort=True, 
                save_format=DFT_STYLE.get_pwdata_format(resource.dft_style), 
                save_path=work_dir, 
                save_name=DFT_STYLE.get_pertub_config(resource.dft_style)
                )

def do_post_pertub(resource: Resource, input_param:InitBulkParam):
    pertub_dir = os.path.join(input_param.root_dir,TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.pertub)
    real_pertub_dir = os.path.join(input_param.root_dir,INIT_BULK.pertub)
    if os.path.exists(pertub_dir):
        link_file(pertub_dir, real_pertub_dir)
    tag = os.path.join(pertub_dir, INIT_BULK.tag_pertub)
    if not os.path.exists(tag):
        write_to_file(tag, "pertub done ", "w")

def pertub_done(input_param:InitBulkParam):
    pertub_dir = os.path.join(input_param.root_dir,TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.pertub)
    tag = os.path.join(pertub_dir, INIT_BULK.tag_pertub)
    return True if os.path.exists(tag) else False

'''
description: 
    get configs:
    the order if pertub/*.configs > *-scaled.config > supercell.config > final.config(relax) > init.config
param {str} super_cell_scale_dir
param {str} relax_dir
param {str} init_config_dirname: the dir made by init_config name
param {str} init_config_path: the init config file path
return {*}
author: wuxingxing
'''
def get_config_files_with_order(super_cell_scale_dir:str, relax_dir:str, init_config_dirname:str, init_config_path:str, pertub_dir:str=None, dft_style:str=None):
    config_list = []
    config_type = INIT_BULK.init
    if pertub_dir is not None:
        config_list = search_files(pertub_dir, "{}/*/*{}".format(init_config_dirname, DFT_STYLE.get_pertub_config(dft_style)))
        config_type = INIT_BULK.pertub
    if len(config_list) == 0:
    # find scaled configs under super_cell_scale dir
        config_list = search_files(super_cell_scale_dir, "{}/*{}".format(init_config_dirname, DFT_STYLE.get_scale_config(dft_style)))
        config_type = INIT_BULK.scale
    if len(config_list) == 0:
    # find super cell config under super_cell_scale dir
        config_list = search_files(super_cell_scale_dir, "{}/{}".format(init_config_dirname, DFT_STYLE.get_super_cell_config(dft_style)))
        config_type = INIT_BULK.super_cell

    if len(config_list) == 0:
    # find relax config from relax dir
        config_list = search_files(relax_dir, "{}/{}".format(init_config_dirname, DFT_STYLE.get_relaxed_config(dft_style)))
        config_type = INIT_BULK.relax

    if len(config_list) == 0:
        # use init config
        config_list = [init_config_path]
        config_type = INIT_BULK.init

    return config_list, config_type


from enum import Enum
from subprocess import Popen, PIPE
import os
import sys
import time
import shutil
import numpy as np
from pwact.utils.constant import EXPLORE_FILE_STRUCTURE
from pwact.active_learning.slurm.slurm_tool import get_jobs
class JobStatus (Enum) :
    unsubmitted = 1 #
    waiting = 2 # PD
    running = 3 # R
    terminated = 4
    finished = 5
    unknown = 100
    resubmit_failed = 6
    submit_limit:int = 3

def get_slurm_sbatch_cmd(job_dir:str, job_name:str):
    cmd = "cd {} && sbatch {}".format(job_dir, job_name)
    return cmd

class SlurmJob(object):
    def __init__(self, job_id=None, status=JobStatus.unsubmitted, user=None, name=None, nodes=None, nodelist=None, partition=None, lmps_tolerance:bool=True) -> None:
        self.job_id = job_id
        self.status = status
        self.user = user
        self.name = name
        self.partition=partition
        self.nodes = nodes
        self.nodelist = nodelist
        self.submit_num = 0
        self.lmps_tolerance = True
        
    def set_cmd(self, script_path:str):
        #such as "sbatch main_MD_test.sh"
        self.slurm_job_run_dir = os.path.dirname(script_path)
        self.slurm_job_name = os.path.basename(script_path)
        slurm_cmd = get_slurm_sbatch_cmd(self.slurm_job_run_dir, self.slurm_job_name)
        self.submit_cmd = slurm_cmd
    
    '''
    description: 
        the job_type could be:
            cp2k/relax, cp2k/scf, cp2k/aimd, pwmat/relax, pwmat/scf, pwmat/aimd, vasp/relax, vasp/scf, vasp/aimd, lammps
    param {*} self
    param {*} tag
    param {str} job_type
    return {*}
    author: wuxingxing
    '''    
    def set_tag(self, tag, job_type:str=None):
        self.job_finish_tag = tag
        if job_type is not None: # use to determine if the lammps md task has terminated due to "ERROR: there are two atoms too close" reason
            self.job_type = job_type.lower()
        else:
            self.job_type = None

    def submit(self):
        # ret = Popen([self.submit_cmd + " " + self.job_script], stdout=PIPE, stderr=PIPE, shell = True)
        ret = Popen([self.submit_cmd], stdout=PIPE, stderr=PIPE, shell = True)
        stdout, stderr = ret.communicate()
        if str(stderr, encoding='ascii') != "":
            raise RuntimeError (stderr)
        job_id = str(stdout, encoding='ascii').replace('\n','').split()[-1]
        self.job_id = job_id
        self.submit_num += 1
        status = self.update_status()
        print ("# job {} submitted!".format(self.job_id))

    def scancel_job(self):
        ret = Popen (["scancel " + self.job_id], shell=True, stdout=PIPE, stderr=PIPE)
        time.sleep(1)
        stdout, stderr = ret.communicate()
        print("scancel job {}".format(self.job_id))
        # print(str(stderr, encoding='ascii'))

    def update_status(self):
        self.status = self.check_status()
        return self.status

    def check_status_no_tag(self):
        ret = Popen (["squeue --job " + self.job_id], shell=True, stdout=PIPE, stderr=PIPE)
        stdout, stderr = ret.communicate()
        if (ret.returncode != 0) :
            if str("Invalid job id specified") in str(stderr, encoding='ascii') :
                return JobStatus.finished
            else :
                print("status command " + "squeue" + " fails to execute")
                print("erro info: " + str(stderr, encoding='ascii'))
                print("return code: " + str(ret.returncode))
                sys.exit ()
        status_line = str(stdout, encoding='ascii').split ('\n')[-2]
        status_word = status_line.split ()[4]
        if   status_word in ["PD","CF","S"] :
            return JobStatus.waiting
        elif status_word in ["R","CG"] :
            return JobStatus.running
        elif status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
            return JobStatus.finished
        elif status_word in ["RH"] : #for job in 'RH' status, scancel the job and return terminated
            self.scancel_job()
            return JobStatus.finished
        else:
            return JobStatus.unknown

    def check_status(self):
        ret = Popen (["squeue --job " + self.job_id], shell=True, stdout=PIPE, stderr=PIPE)
        stdout, stderr = ret.communicate()
        if (ret.returncode != 0) :
            if str("Invalid job id specified") in str(stderr, encoding='ascii') :
                if os.path.exists (self.job_finish_tag) :
                    print("job {} finished: the cmd is {}.".format(self.job_id, self.submit_cmd))
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                print("status command " + "squeue" + " fails to execute")
                print("erro info: " + str(stderr, encoding='ascii'))
                print("return code: " + str(ret.returncode))
                sys.exit ()
        status_line = str(stdout, encoding='ascii').split ('\n')[-2]
        status_word = status_line.split ()[4]
        if   status_word in ["PD","CF","S"] :
            return JobStatus.waiting
        elif status_word in ["R","CG"] :
            return JobStatus.running
        elif status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
            if os.path.exists (self.job_finish_tag) :
                print("job {} finished: the cmd is {}.".format(self.job_id, self.submit_cmd))
                return JobStatus.finished
            else :
                # for lammps md job, if the job stops because of 'ERROR: there are two atoms too close', set the job.status to finished
                if self.job_type is not None and self.job_type == "lammps":
                    end_normal = self.check_lammps_out_file()
                    if end_normal:
                        with open(self.job_finish_tag, 'w') as wf:
                            wf.writelines("Job done!")
                        print("job {} finished: the cmd is {}.".format(self.job_id, self.submit_cmd))
                        return JobStatus.finished
                return JobStatus.terminated
        elif status_word in ["RH"] : #for job in 'RH' status, scancel the job and return terminated
                self.scancel_job()
                if os.path.exists (self.job_finish_tag) :
                    print("job {} finished: the cmd is {}.".format(self.job_id, self.submit_cmd))
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
        else :
            return JobStatus.unknown

    def running_work(self):
        self.submit()
        while True:
            status = self.check_status()
            if (status == JobStatus.waiting) or \
                (status == JobStatus.running):
                time.sleep(10)
            else:
                break
        
        assert(status == JobStatus.finished)
        return status

    def get_slurm_works_dir(self):
        with open(os.path.join(self.slurm_job_run_dir, self.slurm_job_name), 'r') as rf:
            lines = rf.readlines()
        work_dir_list = []
        for line in lines:
            if 'cd ' in line:
                work_dir = line.split()[-1].strip()
                work_dir_list.append(work_dir)
        return work_dir_list

    '''
    description: 
        if the job is md task, and stoped because of 'ERROR: there are two atoms too close' let the task end normally
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def check_lammps_out_file(self):
        # read last line of md.log file
        md_dirs = self.get_slurm_works_dir()
        try:
            for md_dir in md_dirs:
                tag_md_file = os.path.join(md_dir, "tag.md.success")
                md_log = os.path.join(md_dir, "md.log")
                if os.path.exists(tag_md_file):
                    continue

                # check if has error
                with open(md_log, "rb") as file:
                    file.seek(-2, 2)  # 定位到文件末尾前两个字节
                    while file.read(1) != b'\n':  # 逐字节向前查找换行符
                        file.seek(-2, 1)  # 向前移动两个字节
                    last_line = file.readline().decode().strip()  # 读取最后一行并去除换行符和空白字符
                if "Total wall time" in last_line: #md 正常结束
                    with open(tag_md_file, 'w') as wf:
                        wf.writelines("Job Done!")
                    return True

                if os.path.exists(os.path.join(md_dir, EXPLORE_FILE_STRUCTURE.model_devi)):
                    devi = np.loadtxt(os.path.join(md_dir, EXPLORE_FILE_STRUCTURE.model_devi))
                    if self.lmps_tolerance and devi.shape[0] > 0:
                        with open(tag_md_file, 'w') as wf:
                            wf.writelines("Job Done!")
                        return True
                    else:
                        return False
                else: # md运行中非正常结束
                    return False

                # check model_devi.out
                # elif "ERROR: there are two atoms" in last_line:
                #     with open(tag_md_file, 'w') as wf:
                #         wf.writelines("ERROR: there are two atoms too close")
                #     return True
                # else:
                #     return False
            return True
        except Exception as e:
            return False


class Mission(object):
    def __init__(self, mission_id=None) -> None:
        self.mission_id = mission_id
        self.job_list: list[SlurmJob]= []
    
    def add_job(self, job:SlurmJob):
        self.job_list.append(job)

    def pop_job(self, job_id):
        del_job, index = self.get_job(job_id)
        self.job_list.remove(del_job)

    def get_job(self, job_id):
        for i, job in enumerate(self.job_list):
            if job.job_id == job_id:
                return job, i

    def update_job_state(self, job_id, state):
        up_job, index = self.get_job(job_id)
        up_job.status = state
        self.job_list[index] = up_job
    
    def get_running_jobs(self):
        job_list: list[SlurmJob] = []
        for job in self.job_list:
            if (job.status == JobStatus.waiting) or (job.status == JobStatus.running):
                job_list.append(job)
        return job_list

    def move_slurm_log_to_slurm_work_dir(self, slurm_log_dir_source:str):
        for job in self.job_list:
            slurm_log_source = os.path.join(slurm_log_dir_source, "slurm-{}.out".format(job.job_id))
            slurm_job_log_target = os.path.join(os.path.dirname(job.slurm_job_path), os.path.basename(slurm_log_source))
            if os.path.exists(slurm_log_source):
                shutil.move(slurm_log_source, slurm_job_log_target)
                
    '''
    Description: 
    job_finish_tag does not exist means this job running in error 
    param {*} self
    Returns: 
    Author: WU Xingxing
    '''
    def get_error_jobs(self):
        job_list: list[SlurmJob] = []
        for job in self.job_list:
            if os.path.exists(job.job_finish_tag) is False:
                job_list.append(job)
        return job_list
    
    def all_job_finished(self, error_type:str=None):
        error_jobs = self.get_error_jobs()
        if len(error_jobs) >= 1:
            error_log_content = ""
            for error_job in error_jobs:
                error_log_path = os.path.join(error_job.slurm_job_run_dir, "slurm-{}.out".format(error_job.job_id))
                error_log_content += "JOB ERROR! The cmd '{}' failed!\nFor more details on errors, please refer to the following documents:\n"\
                    .format(error_job.submit_cmd)

                slurm_content = "    Slurm script file is {}\n    The slurm log is {}\n"\
                    .format(os.path.join(error_job.slurm_job_run_dir, error_job.slurm_job_name), error_log_path)

                tmp_error = None
                if error_type is not None:
                    work_dirs = error_job.get_slurm_works_dir()
                    if len(work_dirs) > 0:
                        tmp_error = "    Task logs under this slurm job:\n"
                        for _ in work_dirs:
                            job_error_log  = "{}/{}".format(_, error_type)
                            job_finish_tag = "{}/{}".format(_, error_job.job_finish_tag)
                            if os.path.exists(job_error_log) and not os.path.exists(job_finish_tag):
                                tmp_error += "        {}\n".format(job_error_log)

                error_log_content += slurm_content
                if tmp_error is not None:
                    error_log_content += tmp_error
                error_log_content += "\n\n"
            raise Exception(error_log_content)
        return True
    
    def commit_jobs(self):
        for job in self.job_list:
            if job.status == JobStatus.unsubmitted:
                job.submit()
    
    '''
    description: 
        return all job ids, the job id is the slurm job id
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def get_all_job_ids(self):
        job_id_list = []
        for job in self.job_list:
            job_id_list.append(job.job_id)
        return job_id_list
    
    def check_running_job(self):
        while True:
            for job in self.job_list:
                # print(job.status, job.job_id)
                if job.status == JobStatus.resubmit_failed or job.status == JobStatus.finished: # For job resubmitted more than 3 times, do not check again
                    continue
                status = job.check_status()
                self.update_job_state(job.job_id, status)
            # if the job failed, resubmit it until the resubmit time more than 3 times
            self.resubmit_jobs()
            if len(self.get_running_jobs()) == 0:
                break
            time.sleep(10)
        # error_jobs = self.get_error_jobs()
        # if len(error_jobs) > 0:
        #     error_info = "job error: {}".format([_.job_id for _ in error_jobs])
        #     raise Exception(error_info)
        return True
    
    def resubmit_jobs(self):
        for job in self.job_list:
            if job.status == JobStatus.terminated:
                if job.submit_num <= JobStatus.submit_limit.value:
                    print("resubmit job {}: {}, the time is {}\n".format(job.job_id, job.submit_cmd, job.submit_num))
                    job.submit()
                else:
                    job.status = JobStatus.resubmit_failed                    
                
    '''
    Description: 
    after some jobs finished with some jobs terminated, we should try to recover these terminated jobs.
    param {*} self
    Returns: 
    Author: WU Xingxing
    '''
    def re_submmit_terminated_jobs(self):
        error_jobs = self.get_error_jobs()
        if len(error_jobs) == 0:
            return
        self.job_list.clear()
        self.job_list.extend(error_jobs)
        self.reset_job_state()
        self.commit_jobs()
        self.check_running_job()

    def reset_job_state(self):
        for job in self.job_list:
            job.status == JobStatus.unsubmitted

def scancle_job(work_dir:str):
    job_id_list = get_jobs(work_dir)
    print("the job to be scancelled is:")
    print(job_id_list)
    for job_id in job_id_list:
        job = SlurmJob(job_id=job_id)
        status = job.check_status_no_tag()#get status
        if status == JobStatus.waiting or status == JobStatus.running: # is running 
            job.scancel_job()
            # time.sleep(2)
            # status = job.check_status_no_tag()
            # if JobStatus.finished == status:
            #     print("scancel job {} successfully\n\n".format(job_id))
            # else:
            #     print("Scancel job {} failed, Please manually check and cancel this task!\n\n".format(job_id))
    time.sleep(5)
    for job_id in job_id_list:
        job = SlurmJob(job_id=job_id)
        status = job.check_status_no_tag()#get status
        if JobStatus.finished == status:
            print("scancel job {} successfully".format(job_id))
        else:
            print("Scancel job {} failed, Please manually check and cancel this task!\n".format(job_id))

def scancle_byjobid(job_id):
    job = SlurmJob(job_id=job_id)
    status = job.check_status_no_tag()#get status
    if status == JobStatus.waiting or status == JobStatus.running: # is running 
        job.scancel_job()
    time.sleep(5)
    job = SlurmJob(job_id=job_id)
    status = job.check_status_no_tag()#get status
    if JobStatus.finished == status:
        print("scancel job {} successfully".format(job_id))
    else:
        print("Scancel job {} failed, Please manually check and cancel this task!\n".format(job_id))




"""
    dir or label:
    iter.0000/label/scf:
    --------------------/*-scf.job file
    --------------------/*-scf.tag.success file
    --------------------/md.000.sys.000/ dir
    -----------------------------------/md.000.sys.000.t.000.p.000 or md.000.sys.000.t.000 dir
    -------------------------------------------------------------/scf_0/atom.config etot.input pseudo files
    -------------------------------------------------------------/scf_2/atom.config etot.input pseudo files
    -------------------------------------------------------------...
    
    iter.0000/label/result:
    -------------------/summary.txt
    -------------------/md.000.sys.000-mvm, md.000.sys.001-mvm, ...
    
    the content of summary.txt is:
            md.000.sys.000-mvm: image_nums, atom_type
            md.000.sys.001-mvm: image_nums, atom_type
            ...
"""
import os
import glob
import pandas as pd

from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.iter_input import InputParam
from pwact.active_learning.slurm.slurm import SlurmJob, Mission, scancle_job

from pwact.utils.constant import DFT_TYPE, VASP, PWDATA, AL_STRUCTURE, TEMP_STRUCTURE,\
    LABEL_FILE_STRUCTURE, EXPLORE_FILE_STRUCTURE, LAMMPS, SLURM_OUT, DFT_STYLE, PWMAT, INIT_BULK
    
from pwact.utils.slurm_script import get_slurm_job_run_info, split_job_for_group, set_slurm_script_content
from pwact.utils.format_input_output import get_iter_from_iter_name, get_md_sys_template_name
from pwact.utils.file_operation import write_to_file, copy_file, copy_dir, search_files, mv_file, add_postfix_dir, del_dir, del_file_list_by_patten, link_file
from pwact.utils.app_lib.common import link_pseudo_by_atom, set_input_script, check_convergence

from pwact.data_format.configop import extract_pwdata, save_config, get_atom_type
from pwdata import Config

class Labeling(object):
    @staticmethod
    def kill_job(root_dir:str, itername:str):
        label_dir = os.path.join(root_dir, itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.labeling) 
        scf_dir = os.path.join(label_dir, LABEL_FILE_STRUCTURE.scf)
        scancle_job(scf_dir)

    def __init__(self, itername:str, resource: Resource, input_param:InputParam):
        self.itername = itername
        self.iter = get_iter_from_iter_name(self.itername)
        self.resource = resource
        self.input_param = input_param
        
        self.md_job = self.input_param.explore.md_job_list[self.iter]
        
        # train work dir
        self.train_dir = os.path.join(self.input_param.root_dir, self.itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.train)
        self.real_train_dir = os.path.join(self.input_param.root_dir, self.itername, AL_STRUCTURE.train)

        # md work dir
        self.explore_dir = os.path.join(self.input_param.root_dir, itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.explore)
        self.real_explore_dir = os.path.join(self.input_param.root_dir, itername, AL_STRUCTURE.explore)
        self.md_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.md)
        self.select_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.select)
        self.direct_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.direct) 
        self.real_md_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.md)
        self.real_select_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.select)
        self.real_direct_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.direct)
        # labed work dir
        self.label_dir = os.path.join(self.input_param.root_dir, itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.labeling) 
        self.scf_dir = os.path.join(self.label_dir, LABEL_FILE_STRUCTURE.scf)
        self.result_dir = os.path.join(self.label_dir, LABEL_FILE_STRUCTURE.result)

        self.real_label_dir = os.path.join(self.input_param.root_dir, itername, AL_STRUCTURE.labeling) 
        self.real_scf_dir = os.path.join(self.real_label_dir, LABEL_FILE_STRUCTURE.scf)
        self.real_result_dir = os.path.join(self.real_label_dir, LABEL_FILE_STRUCTURE.result)

        self.bigmodel_dir = os.path.join(self.label_dir, LABEL_FILE_STRUCTURE.bigmodel)
        self.real_bigmodel_dir = os.path.join(self.real_label_dir, LABEL_FILE_STRUCTURE.bigmodel)

    '''
    description: 
    the scf work dir file structure is as follow.
    iter.0000/label/scf:
    --------------------/*-scf.job file
    --------------------/*-scf.tag.success file
    --------------------/md.000.sys.000/ 
    -----------------------------------/md.000.sys.000.t.000.p.000 or md.000.sys.000.t.000 
    -------------------------------------------------------------/0-scf/atom.config etot.input pseudo files
    -------------------------------------------------------------/2-scf/atom.config etot.input pseudo files
    -------------------------------------------------------------...
    param {*} self
    return {*}
    author: wuxingxing
    '''        

    def make_scf_work(self):
        candidate = pd.read_csv(os.path.join(self.select_dir, EXPLORE_FILE_STRUCTURE.candidate))
        # make scf work dir
        scf_dir_list = []
        for index, row in candidate.iterrows():
            config_index    = int(row["config_index"])
            sub_md_sys_path = row["file_path"]
            sub_md_sys_name = os.path.basename(sub_md_sys_path)
            md_sys_path     = os.path.dirname(sub_md_sys_path)
            md_sys_name     = os.path.basename(md_sys_path)
            scf_sub_md_sys_path = os.path.join(self.scf_dir, md_sys_name, sub_md_sys_name, "{}-{}".format(config_index, LABEL_FILE_STRUCTURE.scf))
            if not os.path.exists(scf_sub_md_sys_path):
                os.makedirs(scf_sub_md_sys_path)
            tarj_lmp = os.path.join(sub_md_sys_path, EXPLORE_FILE_STRUCTURE.traj, "{}{}".format(config_index, LAMMPS.traj_postfix))
            atom_names = None
            with open(os.path.join(sub_md_sys_path, LAMMPS.atom_type_file), 'r') as rf:
                line = rf.readline()
                atom_names = line.split()
            self.make_scf_file(scf_sub_md_sys_path, tarj_lmp, atom_names)
            scf_dir_list.append(scf_sub_md_sys_path)
        
        self.make_scf_slurm_job_files(scf_dir_list)

    def make_bigmodel_work(self):
        # copy from realdir/direct/select.xyz
        if self.input_param.strategy.direct:
            copy_file(os.path.join(self.real_direct_dir, EXPLORE_FILE_STRUCTURE.select_xyz), 
                os.path.join(self.bigmodel_dir, EXPLORE_FILE_STRUCTURE.select_xyz))
        else:
            # copy trajs to bigmodel_dir and cvt to xyz
            candidate = pd.read_csv(os.path.join(self.select_dir, EXPLORE_FILE_STRUCTURE.candidate))
            # make scf work dir
            image_list = None
            for index, row in candidate.iterrows():
                config_index    = int(row["config_index"])
                sub_md_sys_path = row["file_path"]
                atom_names = None
                with open(os.path.join(sub_md_sys_path, LAMMPS.atom_type_file), 'r') as rf:
                    line = rf.readline()
                    atom_names = line.split()
                if image_list is None:
                    image_list = Config(data_path=os.path.join(sub_md_sys_path, EXPLORE_FILE_STRUCTURE.traj, "{}{}".format(config_index, LAMMPS.traj_postfix)), 
                                        format=PWDATA.lammps_dump, atom_names=atom_names)
                else:
                    image_list.append(Config(data_path=os.path.join(sub_md_sys_path, EXPLORE_FILE_STRUCTURE.traj, "{}{}".format(config_index, LAMMPS.traj_postfix)), 
                                        format=PWDATA.lammps_dump, atom_names=atom_names))
            # cvt_lammps.dump to extxyz
            image_list.to(data_path=self.bigmodel_dir, format=PWDATA.extxyz, data_name="{}".format(EXPLORE_FILE_STRUCTURE.select_xyz))
        # copy bigmodelscript
        copy_file(self.input_param.scf.bigmodel_script, os.path.join(self.bigmodel_dir, os.path.basename(self.input_param.scf.bigmodel_script)))
        # make slrum file
        self.make_bigmodel_slurm_job_files([self.bigmodel_dir])

    def back_label(self):
        if self.input_param.scf.dft_style == DFT_STYLE.bigmodel:
            slurm_remain, slurm_success = get_slurm_job_run_info(self.real_bigmodel_dir, \
                job_patten="*-{}".format(LABEL_FILE_STRUCTURE.bigmodel_job), \
                tag_patten="*-{}".format(LABEL_FILE_STRUCTURE.bigmodel_tag),
                for_back = True)
            slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False        
        else:
            slurm_remain, slurm_success = get_slurm_job_run_info(self.real_scf_dir, \
                job_patten="*-{}".format(LABEL_FILE_STRUCTURE.scf_job), \
                tag_patten="*-{}".format(LABEL_FILE_STRUCTURE.scf_tag))
            slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False
        
        if slurm_done:
            # bk and do new job
            target_bk_file = add_postfix_dir(self.real_label_dir, postfix_str="bk")
            mv_file(self.real_label_dir, target_bk_file)
            # if the temp_work_dir/label exists, delete the train dir
            if os.path.exists(self.label_dir):
                del_dir(self.label_dir)

    def do_scf_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.scf_dir, \
            job_patten="*-{}".format(LABEL_FILE_STRUCTURE.scf_job), \
            tag_patten="*-{}".format(LABEL_FILE_STRUCTURE.scf_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False
        if slurm_done is False:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these SCF Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), LABEL_FILE_STRUCTURE.scf_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished(error_type=SLURM_OUT.dft_out)

    def do_bigmodel_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.bigmodel_dir, \
            job_patten="*-{}".format(LABEL_FILE_STRUCTURE.bigmodel_job), \
            tag_patten="*-{}".format(LABEL_FILE_STRUCTURE.bigmodel_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False
        if slurm_done is False:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run bigModel Job:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), LABEL_FILE_STRUCTURE.bigmodel_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished()

    def make_scf_file(self, scf_dir:str, tarj_lmp:str, atom_names:list[str]=None):
        config_index = os.path.basename(tarj_lmp).split('.')[0]
        if DFT_STYLE.vasp == self.resource.dft_style: # when do scf, the vasp input file name is 'POSCAR'
            save_name = VASP.poscar
        else:
            save_name="{}{}".format(config_index, DFT_STYLE.get_normal_config(self.resource.dft_style))# for cp2k this param will be set as coord.xzy
        target_config = save_config(config=tarj_lmp,
                                    input_format=PWDATA.lammps_dump,
                                    wrap = False, 
                                    direct = True, 
                                    sort = True, 
                                    save_name = save_name,
                                    save_format=DFT_STYLE.get_pwdata_format(dft_style=self.resource.dft_style, is_cp2k_coord=True),
                                    save_path=scf_dir, 
                                    atom_names=atom_names)

        #2.
        atom_type_list = atom_names
        #1. set pseudo files
        if not self.input_param.scf.use_dftb:
            pseudo_names = link_pseudo_by_atom(
                    pseudo_list = self.input_param.scf.pseudo, 
                    target_dir = scf_dir, 
                    atom_order = atom_type_list, 
                    dft_style = self.resource.dft_style,
                    basis_set_file  =self.input_param.scf.basis_set_file,
                    potential_file  =self.input_param.scf.potential_file
                    )
        else:
            # link in.skf path to aimd dir
            target_dir = os.path.join(scf_dir, PWMAT.in_skf)
            if self.input_param.scf.in_skf is not None:
                link_file(self.input_param.scf.in_skf, target_dir)
            pseudo_names = []
        #2. make etot.input file
        set_input_script(
            input_file=self.input_param.scf.scf_input_list[0].input_file,
            config=target_config,
            dft_style=self.resource.dft_style,
            kspacing=self.input_param.scf.scf_input_list[0].kspacing, 
            flag_symm=self.input_param.scf.scf_input_list[0].flag_symm, 
            save_dir=scf_dir,
            pseudo_names=pseudo_names,
            gaussian_base_param=self.input_param.scf.gaussian_base_param,# these for cp2k
            is_scf = True
        )
        
    def make_scf_slurm_job_files(self, scf_sub_list:list[str]):
        del_file_list_by_patten(self.scf_dir, "*{}".format(LABEL_FILE_STRUCTURE.scf_job))
        group_list = split_job_for_group(self.resource.dft_resource.group_size, scf_sub_list, self.resource.dft_resource.parallel_num)
        
        for group_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            
            jobname = "scf{}".format(group_index)
            tag_name = "{}-{}".format(group_index, LABEL_FILE_STRUCTURE.scf_tag)
            tag = os.path.join(self.scf_dir, tag_name)
            run_cmd = self.resource.dft_resource.command
            # if self.resource.dft_resource.gpu_per_node > 0:
            #     run_cmd = "mpirun -np {} PWmat > {}".format(self.resource.dft_resource.gpu_per_node, SLURM_OUT.md_out)
            # else:
            #     raise Exception("ERROR! the cpu version of pwmat not support yet!")
            group_slurm_script = set_slurm_script_content(gpu_per_node=self.resource.dft_resource.gpu_per_node, 
                number_node = self.resource.dft_resource.number_node, 
                cpu_per_node = self.resource.dft_resource.cpu_per_node,
                queue_name = self.resource.dft_resource.queue_name,
                custom_flags = self.resource.dft_resource.custom_flags,
                env_script = self.resource.dft_resource.env_script,
                job_name = jobname,
                run_cmd_template = run_cmd,
                group = group,
                job_tag = tag,
                task_tag = LABEL_FILE_STRUCTURE.scf_tag, 
                task_tag_faild = LABEL_FILE_STRUCTURE.scf_tag_failed,
                parallel_num=self.resource.dft_resource.parallel_num,
                check_type=self.resource.dft_style
                )
            slurm_script_name = "{}-{}".format(group_index, LABEL_FILE_STRUCTURE.scf_job)
            slurm_job_file = os.path.join(self.scf_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")


    def make_bigmodel_slurm_job_files(self, scf_sub_list:list[str]):
        del_file_list_by_patten(self.bigmodel_dir, "*{}".format(LABEL_FILE_STRUCTURE.scf_job))
        group_list = split_job_for_group(1, scf_sub_list, 1)
        
        for group_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            
            jobname = "bigmodel{}".format(group_index)
            tag_name = "{}-{}".format(group_index, LABEL_FILE_STRUCTURE.bigmodel_tag)
            tag = os.path.join(self.bigmodel_dir, tag_name)
            run_cmd = self.resource.dft_resource.command
            # if self.resource.dft_resource.gpu_per_node > 0:
            #     run_cmd = "mpirun -np {} PWmat > {}".format(self.resource.dft_resource.gpu_per_node, SLURM_OUT.md_out)
            # else:
            #     raise Exception("ERROR! the cpu version of pwmat not support yet!")
            group_slurm_script = set_slurm_script_content(gpu_per_node=self.resource.dft_resource.gpu_per_node, 
                number_node = self.resource.dft_resource.number_node, 
                cpu_per_node = self.resource.dft_resource.cpu_per_node,
                queue_name = self.resource.dft_resource.queue_name,
                custom_flags = self.resource.dft_resource.custom_flags,
                env_script = self.resource.dft_resource.env_script,
                job_name = jobname,
                run_cmd_template = run_cmd,
                group = group,
                job_tag = tag,
                task_tag = LABEL_FILE_STRUCTURE.bigmodel_tag, 
                task_tag_faild = LABEL_FILE_STRUCTURE.bigmodel_tag_failed,
                parallel_num=self.resource.dft_resource.parallel_num,
                check_type=self.resource.dft_style
                )
            slurm_script_name = "{}-{}".format(group_index, LABEL_FILE_STRUCTURE.bigmodel_job)
            slurm_job_file = os.path.join(self.bigmodel_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    '''
    description: 
    collecte OUT.MLMD to mvm-
    param {*} self
    return {*}
    author: wuxingxing
    '''         
    def collect_scf_configs(self):
        md_sys_mlmd = []
        md_sys_dir_list = search_files(self.scf_dir, get_md_sys_template_name())
        for md_sys_dir in md_sys_dir_list:
            sub_md_sys_dir_list =search_files(md_sys_dir, get_md_sys_template_name())
            for sub_md_sys in sub_md_sys_dir_list:
                out_mlmd_list =search_files(sub_md_sys, "*-{}/{}".format(LABEL_FILE_STRUCTURE.scf, DFT_STYLE.get_scf_config(self.resource.dft_style)))
                # do a sorted?
                md_sys_mlmd.extend(out_mlmd_list)
        return md_sys_mlmd

    def get_aimd_list(self):
        aimd_list = search_files(self.scf_dir, "{}/{}".format(get_md_sys_template_name(), DFT_STYLE.get_aimd_config(self.resource.dft_style)))
        return sorted(aimd_list)
    
    '''
    description: 
        if reserve_work is True
            reserve all temp work dir
        else:
            if reserve_scf_files is True, copy scf and result files
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def do_post_labeling(self):
        # collect scf files
        scf_dirs = search_files(self.scf_dir, "{}/{}/*{}".format(get_md_sys_template_name(),get_md_sys_template_name(), LABEL_FILE_STRUCTURE.scf))
        for scf_dir in scf_dirs:
            if self.input_param.reserve_scf_files:
                target_scf_dir = scf_dir.replace(TEMP_STRUCTURE.tmp_run_iter_dir, "") 
                copy_dir(scf_dir, target_scf_dir)
            else:
                scf_files = os.listdir(scf_dir)
                for scf_file in scf_files:
                    scf_file_path = os.path.join(scf_dir, scf_file)
                    if scf_file.lower() in DFT_STYLE.get_scf_reserve_list(self.resource.dft_style) \
                        or "atom.config" in scf_file.lower() :# for the input natom.config
                        copy_file(scf_file_path, scf_file_path.replace(TEMP_STRUCTURE.tmp_run_iter_dir, ""))

        # scf files to pwdata format
        scf_configs = self.collect_scf_configs()
        cvg_files, uncvg_files, cvg_infos, cvg_detail_infos = check_convergence(scf_configs, self.resource.dft_style)
        if len(cvg_files) > 0:
            extract_pwdata(input_data_list=cvg_files,
                    intput_data_format =DFT_STYLE.get_format_by_postfix(os.path.basename(scf_configs[0])),
                    save_data_path =self.result_dir,
                    save_data_name = INIT_BULK.get_save_format(self.input_param.data_format),
                    save_data_format = self.input_param.data_format,
                    data_shuffle     =self.input_param.train.data_shuffle
            )
            # copy to main dir
            copy_dir(self.result_dir, self.real_result_dir)
        print(cvg_detail_infos)
        write_to_file(os.path.join(self.input_param.root_dir, EXPLORE_FILE_STRUCTURE.iter_select_file), cvg_detail_infos, mode='a')

    def do_post_bigmodel(self):
        # copy the bigmodel labeled.xyz to result
        if self.input_param.data_format == PWDATA.extxyz:
            copy_file(os.path.join(self.bigmodel_dir, LABEL_FILE_STRUCTURE.train_xyz), os.path.join(self.result_dir, LABEL_FILE_STRUCTURE.train_xyz))
        else:
            images = Config(data_path=os.path.join(self.bigmodel_dir, LABEL_FILE_STRUCTURE.train_xyz), format=PWDATA.extxyz)
            images.to(data_path=self.result_dir, format=PWDATA.pwmlff_npy)
        # copy bigmodel dir to real dir
        copy_dir(self.bigmodel_dir, self.real_bigmodel_dir)
        copy_dir(self.result_dir, self.real_result_dir)
        # del slurm logs and tags
        del_file_list_by_patten(self.real_bigmodel_dir, "slurm-*")


from pwact.utils.constant import DFT_STYLE, VASP, PWMAT, CP2K, \
    ELEMENTTABLE_2, ELEMENTTABLE
from pwact.utils.file_operation import link_file, merge_files_to_one, write_to_file, copy_file, file_read_last_line, del_file
from pwact.utils.app_lib.pwmat import set_etot_input_by_file
import os
from pwact.data_format.configop import get_atom_type, read_cp2k_xyz
from pwact.utils.app_lib.cp2k import make_cp2k_input, make_cp2k_input_from_external
'''
description: 
    this script for pwmat, vasp, ... common operation
author: wuxingxing
'''

'''
description: 
    link the pseudo files to target dir
    for pwmat pseduo files, link each file to target dir
    for vasp pseduo files, merge files to one potcar file according to atom_order, then write to target dir
    for cp2k:
        copy the basis_set_file, potential_file_name to target dir
        
param {list} pseudo_list
param {str} target_dir
param {list} atom_order
param {str} dft_style
return {*}
author: wuxingxing
'''
def link_pseudo_by_atom(
        pseudo_list:list, 
        target_dir:str, 
        atom_order:list[str], 
        dft_style:str,
        basis_set_file:str=None,
        potential_file:str=None
    ):

    pseudo_find = []
    if dft_style == DFT_STYLE.pwmat:
        for atom_name in atom_order:
            for pseudo in pseudo_list:
                pseudo_path = pseudo[0]
                pseudo_name = os.path.basename(pseudo_path)
                atom_type = pseudo[1]
                if atom_name == atom_type:
                    link_file(pseudo_path, os.path.join(target_dir, pseudo_name))
                    pseudo_find.append(pseudo_path)
                    break
        # assert len(pseudo_find) == len(atom_order), "the pwmat pseudo files {} not same as atom type '{}'".format(pseudo_find, atom_order)
        if basis_set_file is not None and potential_file is not None: # these 2 files for pwmat gaussian base
            link_file(basis_set_file, os.path.join(target_dir, os.path.basename(basis_set_file)))
            link_file(potential_file, os.path.join(target_dir, os.path.basename(potential_file)))

    elif dft_style == DFT_STYLE.vasp:
        # merge file to where? to save dir
        for atom_name in atom_order:
            for pseudo in pseudo_list:
                pseudo_path = pseudo[0]
                atom_type = pseudo[1]
                if atom_name == atom_type:
                    pseudo_find.append(pseudo_path)
        assert len(pseudo_find) == len(atom_order), "the vasp pseudo files {} not same as atom type '{}'".format(pseudo_find, atom_order)
        target_pseudo_file = os.path.join(target_dir, VASP.potcar)
        merge_files_to_one(pseudo_find, target_pseudo_file)
    
    elif dft_style == DFT_STYLE.cp2k:
        link_file(basis_set_file, os.path.join(target_dir, os.path.basename(basis_set_file)))
        link_file(potential_file, os.path.join(target_dir, os.path.basename(potential_file)))

    else:
        pass
    
    return [os.path.basename(_) for _ in pseudo_find] # this return is for pwmat etot.input for IN.PSP set

'''
description: 
    get atom type in pseudo file
param {str} pseduo_path
return {*}
author: wuxingxing
'''
def get_vasp_pseudo_atom_type(pseduo_path:str):
    with open(pseduo_path, "r") as rf:
        line = rf.readline()
    atom_type = line.split()[1].split('_')[0]
    return atom_type


'''
description: 
    set input script, for pwmat is etot.input file, for vasp is INCAR file
return {*}
author: wuxingxing
'''
def set_input_script(            
    input_file:str=None,    # pwmat is etot.input, vasp is INCAR
    config:str=None, # the format is same as dft_style
    kspacing:float=None, 
    flag_symm:int=None, 
    dft_style:str=None,
    save_dir:str=None,
    pseudo_names:list[str]=None,
    is_scf = False, # if is_scf, the pwmat etot.input will set the 'out.mlmd = T'
    gaussian_base_param = None
    # basis_set_file_name=None,
    # potential_file_name=None,
    # basis_set_list=None,
    # potential_list=None
    # xc_functional=None,
    # potential=None,
    # basis_set=None
    ):
    if dft_style == DFT_STYLE.pwmat:
        target_file = os.path.join(save_dir, PWMAT.etot_input)
        is_skf_file = os.path.exists(os.path.join(save_dir, PWMAT.in_skf))
        script = set_etot_input_by_file(    #do pwmat etot.input content check
            etot_input_file=input_file, 
            atom_config=config, 
            kspacing=kspacing, 
            flag_symm=flag_symm, 
            pseudo_names=pseudo_names,
            is_scf = is_scf,
            is_skf_file = is_skf_file,
            gaussian_base_param=gaussian_base_param
            )
        write_to_file(target_file, script, "w")
    elif dft_style == DFT_STYLE.vasp:
        #do pwmat INCAR content check, not realized yet.
        target_file = os.path.join(save_dir, VASP.incar)
        copy_file(input_file, target_file)
    
    elif dft_style == DFT_STYLE.cp2k: # for cp2k is cp2k.inp file
        target_file = os.path.join(save_dir, CP2K.cp2k_inp)
        cell = file_read_last_line(os.path.join(save_dir, CP2K.cell_txt), type_name="float")
        del_file(os.path.join(save_dir, CP2K.cell_txt))
        # inp file, cell cood add to inp file
        # set kind_dict

        script = make_cp2k_input_from_external(
            cell=cell,
            coord_file = config,
            exinput_path=input_file,
            gaussian_base_param = gaussian_base_param
            )
        write_to_file(target_file, script, "w")

        # atom_type_name, atom_names, coord = read_cp2k_xyz(target_file) 
        # make_cp2k_input(
        #     cell=image.lattice,
        #     atom_names= atom_type_names, 
        #     atom_types=image.atom_type,
        #     coord_list=image.position,
        #     basis_set_file_name=basis_set_file_name,
        #     potential_file_name=potential_file_name,
        #     # xc_functional=potential_file_name,
        #     # potential=potential,
        #     # basis_set=basis_set
        #     )

    else:
        pass

def is_convergence(file_path, format):
    def _is_cvg_vasp(file_path:str):
        with open(file_path, 'r') as rf:
            outcar_contents = rf.readlines()
        nelm = None
        ediff = None
        for idx, ii in enumerate(outcar_contents):
            if 'NELM   =' in ii:
                nelm = int(ii.split()[2][:-1])
            if 'EDIFF = ' in ii:
                ediff = float(ii.split()[-1])
        
        with open(os.path.join(os.path.dirname(os.path.abspath(file_path)), "OSZICAR"), 'r') as rf:
            oszi_contents = rf.readlines()
        _split = oszi_contents[-2].split()
        real_nelm = int(_split[1])
        real_ediff1 = abs(float(_split[3]))
        real_ediff2 = abs(float(_split[4]))

        if real_nelm < nelm:
            return True
        elif real_ediff1 <= ediff and real_ediff2 <=ediff:
            return True
        else:
            False

    def _is_cvg_pwmat(file_path:str):
        with open(os.path.join(os.path.dirname(os.path.abspath(file_path)), "REPORT"), 'r') as rf:
            report_contents = rf.readlines()
        e_error   = None
        rho_error = None
        etot_idx = -1
        drho_idx = -1
        for idx, ii in enumerate(report_contents):
            if e_error is None and 'E_ERROR   =' in ii:
                e_error = abs(float(ii.split()[-1]))
            if rho_error is None and 'RHO_ERROR =' in ii:
                rho_error = abs(float(ii.split()[-1]))
            if 'E_tot(eV)            =' in ii:
                etot_idx = idx
            if 'dv_ave, drho_tot     =' in ii:
                drho_idx = idx
            if 'niter reached' in ii:
                break
            elif 'ending_scf_reason = tol' in ii:
                return True

        if e_error >= abs(float(report_contents[etot_idx].split()[-1])) or \
            rho_error >= abs(float(report_contents[drho_idx].split()[-1])):
            return True
        return False
    
    def _is_cvg_cp2k(file_path:str):
        with open(os.path.join(os.path.dirname(os.path.abspath(file_path)), "dft.log"), 'r') as rf:
            report_contents = rf.readlines()
        for idx, ii in enumerate(report_contents):
            if 'SCF run NOT converged' in ii:
                return False
        return True

    if format == DFT_STYLE.vasp:
        return _is_cvg_vasp(file_path)
    elif format == DFT_STYLE.pwmat:
        return _is_cvg_pwmat(file_path)
    elif format == DFT_STYLE.cp2k:
        return _is_cvg_cp2k(file_path)
    else: # for other format
        return True

def check_convergence(file_path:list[str], format:str):
    cvg_files = []
    uncvg_files = []
    cvg_infos = ""
    cvg_detail_infos=""
    for file in file_path:
        if is_convergence(file, format):
            cvg_files.append(file)
        else:
            uncvg_files.append(file)
    cvg_infos += "Number of converged files: {}, number of non-converged files: {}\n".format(len(cvg_files), len(uncvg_files))
    cvg_detail_infos += cvg_infos
    if len(uncvg_files) > 0:
        cvg_detail_infos += "List of non-converged files:\n{}".format("\n".join(uncvg_files))
    return cvg_files, uncvg_files, cvg_infos, cvg_detail_infos
    


import os
import numpy as np

'''
description: 
 this NEP params default value is the same as 
    https://gpumd.org/nep/input_files/nep_in.html#nep-in

return {*}
author: wuxingxing
'''

element_table = ['', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl',
                 'Ar',
                 'K', 'Ca',
                 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',
                 'Rb', 'Sr', 'Y',
                 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs',
                 'Ba', 'La',
                 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta',
                 'W', 'Re',
                 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
                 'U', 'Np',
                 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr']

class NepParam(object):
    '''
    description: 
        if nep.in file provided by user, prioritize using it
        elif nep.in file not exisited, use params set in user input json file
        else use default params
    param {*} self
    param {*} nep_dict
    param {str} nep_in_file
    param {list} type_list
    param {list} type_list_weight
    param {str} nep_save_file
    return {*}
    author: wuxingxing
    '''    
    def __init__(self) -> None:
        self.nep_in_file = None
        self.nep_txt_file = None

    '''
    description: 
        extract nep params from nep.in file
    param {*} self
    param {*} nep_in_file
    param {list} type_list
    return {*}
    author: wuxingxing
    '''
    def set_nep_param_from_nep_in(self, nep_in_file, type_list:list[int]=[]) -> None:
        nep_file_dict = {}
        self.nep_in_file = nep_in_file
        if nep_in_file is not None:
            if not os.path.exists(nep_in_file):
                raise Exception("ERROR! the nep.in file is not exist, please check the file: {}".format(nep_in_file))
            nep_file_dict = self.read_nep_param_from_nep_file(nep_in_file)
            
        self.version = self.get_parameter("version", nep_file_dict, 4) # select between NEP2, NEP3, and NEP4
        type_str = self.set_atom_type(type_list)
        self.type = self.get_parameter("type", nep_file_dict, type_str) # number of atom types and list of chemical species
        self.type_num = len(self.type.split()[1:])
        type_list_weight_default = [1.0 for _ in range(0, self.type_num)]
        self.type_weight = self.get_parameter("type_weight", nep_file_dict, type_list_weight_default) # force weights for different atom types
        self.model_type = self.get_parameter("model_type", nep_file_dict, 0) # select to train potential 0, dipole 1, or polarizability 2
        self.prediction = self.get_parameter("prediction", nep_file_dict, 0) # select between training and prediction (inference)
        self.zbl = self.get_parameter("zbl", nep_file_dict, None) # outer cutoff for the universal ZBL potential [Ziegler1985]
        
        self.cutoff = self.get_parameter("cutoff", nep_file_dict, [6, 6], out_format=4) # radial () and angular () cutoffs # use dp rcut, default to 6
        self.n_max = self.get_parameter("n_max", nep_file_dict, [4, 4], out_format=2) # size of radial () and angular () basis
        if len(self.n_max) != 2:
            raise Exception("the input 'n_max' should has 2 values, such as [4, 4]")
        self.basis_size = self.get_parameter("basis_size", nep_file_dict, [12, 12], out_format=2) # number of radial () and angular () basis functions
        if len(self.basis_size) != 2:
            raise Exception("the input 'basis_size' should has 2 values, such as [12, 12]")
        self.l_max = self.get_parameter("l_max", nep_file_dict, [4, 2, 1], out_format=2) # expansion order for angular terms
        if len(self.l_max) != 3 or (self.l_max[0] != 4) or (self.l_max[1] != 2) or (self.l_max[2] > 1) :
            error_log = "the input 'l_max' should has 3 values. The values should be [4, 2, 0] or [4, 2, 1]. The last num '1', means use 5 body features.\n"
            raise Exception(error_log)
        if self.l_max[2] != 0 and self.l_max[2] != 1:
            error_log = "the input 'l_max' should has 3 values. The values should be [4, 2, 0] or [4, 2, 1]. The last num '1', means use 5 body features, '0' means not use 5 body features\n"
            raise Exception(error_log)
        self.neuron = self.get_parameter("neuron", nep_file_dict, 100, out_format=1) # number of neurons in the hidden layer
        self.neuron = [self.neuron, 1]
        self.lambda_1 = self.get_parameter("lambda_1", nep_file_dict, -1, out_format=3) # weight of regularization term
        if self.lambda_1 != -1.0 and self.lambda_1 < 0:
            raise Exception("ERROR! the lambda_1 should >= 0 or lambda_1 = -1 for automatically determined in training!")

        self.lambda_2 = self.get_parameter("lambda_2", nep_file_dict, -1, out_format=3) # weight of norm regularization term
        if self.lambda_2 != -1.0 and self.lambda_2 < 0:
            raise Exception("ERROR! the lambda_2 should >= 0 or lambda_2 = -1 for automatically determined in training!")

        self.lambda_ei = self.get_parameter("lambda_ei", nep_file_dict, 1.0, out_format=3) # weight of energy loss term
        self.lambda_eg = self.get_parameter("lambda_eg", nep_file_dict, 0.1, out_format=3) # weight of energy loss term
        self.lambda_e = self.get_parameter("lambda_e",   nep_file_dict, 1.0, out_format=3) # weight of energy loss term
        self.lambda_f = self.get_parameter("lambda_f",   nep_file_dict, 1.0, out_format=3) # weight of force loss term
        self.lambda_v = self.get_parameter("lambda_v",   nep_file_dict, 0.1, out_format=3) # weight of virial loss term
        self.force_delta = self.get_parameter("force_delta", nep_file_dict, None, out_format=3) # bias term that can be used to make smaller forces more accurate
        self.batch =       self.get_parameter("batch",      nep_file_dict, 1000, out_format=1) # batch size for training
        self.population =  self.get_parameter("population", nep_file_dict, 50, out_format=1) # population size used in the SNES algorithm [Schaul2011]
        self.generation =  self.get_parameter("generation", nep_file_dict, 100000, out_format=1) # number of generations used by the SNES algorithm [Schaul2011]
        self.set_feature_params()

    '''
    description: 
    recover model from nep.txt file
    for nep.txt in gpumd, the bias is '-' op,  but in pwmlff, the bias is '+' op
    param {str} nep_txt_file
    return {*}
    author: wuxingxing
    '''    
    def set_nep_nn_c_param_from_nep_txt(self, nep_txt_file:str):
        self.nep_txt_file = nep_txt_file
        with open(nep_txt_file, "r") as rf:
            lines =rf.readlines()

        line_num = len(lines)

        line_1 = lines[0].split()
        version, type_num, type_list = line_1[0], int(line_1[1]), line_1[2:]
        self.type_num = type_num
        self.zbl = None
        use_zbl = False
        self.use_fixed_zbl = False
        if "zbl" in lines[1]:
            zbl_line = lines[1].split()
            if float(zbl_line[1]) < 1e-04 and float(zbl_line[2]) < 1e-04:
                self.use_fixed_zbl = True
                raise Exception("ERROR! The fixed zbl is not supported!")
            else:
                self.zbl = float(zbl_line[2])
                # print("use zbl: {}".format(self.zbl))
                # if self.zbl[0] > self.zbl[1] or self.zbl[0] > 2.5 :
                #     raise Exception("ERROR! the inner should smaller than outer of zbl in nep.txt! And both of them should be smaller than 2.5!\n The inner = 0.5*outer should be a reasonable choice. ")
                # elif self.zbl[1] > 2.5:
                #     print("Warning! the input zbl outer param should smaller than 2.5! Automatically adjust outer to 2.5!")
                #     self.zbl[1] = 2.5
                lines[1:-1] = lines[2:]
                
            use_zbl = True
            
        cutoffs = lines[1].split()
        self.cutoff = [float(cutoffs[1]), float(cutoffs[2])]

        n_maxs =  lines[2].split() 
        self.n_max = [int(n_maxs[1]), int(n_maxs[2])]

        basis_size =  lines[3].split() 
        self.basis_size = [int(basis_size[1]),int(basis_size[2])]

        l_maxs =  lines[4].split() 
        self.l_max = [int(l_maxs[1]), int(l_maxs[2]), int(l_maxs[3])]

        anns = lines[5].split()
        ann_num = int(anns[1])
        self.neuron = [ann_num, 1]
        self.set_feature_params()

        #num_w0 
        self.model_wb = []
        start_index = 6
        w0_num = self.feature_nums*ann_num
        b0_num = ann_num
        ann_nums= (w0_num + b0_num*2) * self.type_num + self.type_num
        need_line = 6 + ann_nums + self.c_num + self.feature_nums
        if use_zbl:
            if self.use_fixed_zbl:
                need_line += 1 + (self.type_num+1)*self.type_num/2
            else:
                need_line += 1
        is_gpumd_nep = False if need_line == line_num else True
        for i in range(0, self.type_num):
            w0 = np.array([float(_) for _ in lines[start_index        : start_index+w0_num]]).reshape(ann_num, self.feature_nums).transpose(1,0)
            start_index = start_index + w0_num
            self.model_wb.append(w0)
            b0 = np.array([-float(_) for _ in lines[start_index : start_index + b0_num]]).reshape(1,b0_num)
            start_index = start_index + b0_num
            self.model_wb.append(b0)
            w1 = np.array([float(_) for _ in lines[start_index : start_index + b0_num]]).reshape(b0_num, 1)
            start_index = start_index + b0_num
            self.model_wb.append(w1)

        if is_gpumd_nep:
            print("The nep.txt file is from GPUMD")
            common_bias = -float(lines[start_index])
            self.bias_lastlayer = np.array([common_bias for _ in range(0, self.type_num)])
            start_index = start_index + 1
        else:
            print("The nep.txt file is from PWMLFF")
            self.bias_lastlayer = np.array([-float(_) for _ in lines[start_index : start_index + self.type_num]])
            start_index = start_index + self.type_num

        self.c2_param = np.array([float(_) for _ in lines[start_index:start_index + self.two_c_num]]).reshape(self.n_max[0]+1, self.basis_size[0]+1, self.type_num, self.type_num).transpose(2, 3, 0, 1)
        start_index = start_index + self.two_c_num
        self.c3_param = np.array([float(_) for _ in lines[start_index:start_index + self.three_c_num]]).reshape(self.n_max[1]+1, self.basis_size[1]+1, self.type_num, self.type_num).transpose(2, 3, 0, 1)
        start_index = start_index + self.three_c_num
        self.q_scaler = np.array([float(_) for _ in lines[start_index:start_index + self.feature_nums]])
        start_index += self.feature_nums
        if use_zbl:
            if start_index + 1 != line_num:
                raise Exception("extract the nep.txt {} error! the params need {} but has {}\n".format(nep_txt_file, start_index + 1, len(lines)))
        else:
            if start_index != line_num:
                raise Exception("extract the nep.txt {} error! the params need {} but has {}\n".format(nep_txt_file, start_index, len(lines)))

    '''
    description: 
    extract nep params from input json file
    param {*} self
    param {dict} json_dict
    param {list} type_list
    return {*}
    author: wuxingxing
    '''
    def set_nep_param_from_json(self, json_dict:dict, type_list:list[int]=[]):
        model_dict = self.get_parameter("model", json_dict, {})
        descriptor_dict = self.get_parameter("descriptor", model_dict, {})
        # optimizer_dict = self.get_parameter("optimizer", json_dict, {})

        self.version = 4 # select between NEP2, NEP3, and NEP4
        type_str = self.set_atom_type(type_list)
        self.type = type_str # number of atom types and list of chemical species
        self.type_num = len(type_list)
        type_list_weight_default = [1.0 for _ in range(0, self.type_num)]
        self.type_weight = self.get_parameter("type_weight", descriptor_dict, type_list_weight_default) # force weights for different atom types
        self.model_type = 0 # select to train potential 0, dipole 1, or polarizability 2
        self.prediction = 0 # select between training and prediction (inference)
        self.zbl = self.get_parameter("zbl", descriptor_dict, None)
        if self.zbl is not None:
            if self.zbl > 2.5 or self.zbl < 1.0:
                raise Exception("ERROR! the 'zbl' in json file should be between 1.0 and 2.5")
        self.use_fixed_zbl = False

        self.cutoff = self.get_parameter("cutoff", descriptor_dict, [6.0, 6.0]) # radial () and angular () cutoffs # use dp rcut, default to 6
        self.n_max = self.get_parameter("n_max", descriptor_dict, [4, 4]) # size of radial () and angular () basis
        if len(self.n_max) != 2:
            raise Exception("the input 'n_max' should has 2 values, such as [4, 4]")
        self.basis_size = self.get_parameter("basis_size", descriptor_dict, [12, 12]) # number of radial () and angular () basis functions
        if len(self.basis_size) != 2:
            raise Exception("the input 'basis_size' should has 2 values, such as [12, 12]")
        self.l_max = self.get_parameter("l_max", descriptor_dict, [4, 2, 1]) # expansion order for angular terms
        if len(self.l_max) != 3 :
            error_log = "the input 'l_max' should has 3 values. The values should be [4, 0, 0] (only use three body features), [4, 2, 0] (use 3 and 4 body features) or [4, 2, 1] (use 3,4,5 body features).\n"
            raise Exception(error_log)
        
        if (self.l_max[2] in [0, 1]) is False or (self.l_max[1] in [0, 2]) is False:
            error_log = "the input 'l_max' should has 3 values. The values should be [4, 0, 0] (only use three body features), [4, 2, 0] (use 3 and 4 body features) or [4, 2, 1] (use 3,4,5 body features).\n"
            raise Exception(error_log)
        if "fitting_net" in model_dict.keys():
            self.neuron = self.get_parameter("network_size", model_dict["fitting_net"], [40]) # number of neurons in the hidden layer
            if not isinstance(self.neuron, list):
                self.neuron = [self.neuron]
        else:
            self.neuron = [40]
        if self.neuron[-1] != 1:
            self.neuron.append(1) # output layer of fitting net
        self.set_feature_params()

    def set_feature_params(self):
        # features
        self.two_feat_num = self.n_max[0]+1
        self.three_feat_num = (self.n_max[1]+1)*self.l_max[0]
        self.four_feat_num = (self.n_max[1]+1) if self.l_max[1] != 0 else 0
        self.five_feat_num = (self.n_max[1]+1) if self.l_max[2] != 0 else 0
        self.feature_nums = self.two_feat_num + self.three_feat_num + self.four_feat_num + self.five_feat_num
        # c params, the 4-body and 5-body use the same c param of 3-body, their N_base_a the same
        self.two_c_num = self.type_num*self.type_num*(self.n_max[0]+1)*(self.basis_size[0]+1)
        self.three_c_num = self.type_num*self.type_num*(self.n_max[1]+1)*(self.basis_size[1]+1)
        self.c_num = self.two_c_num + self.three_c_num

        self.c2_param = None
        self.c3_param = None
        self.q_scaler = None
        self.model_wb = None

    '''
    read params from nep.txt
    description: 
    param {*} self
    param {*} file_path
    return {*}
    author: wuxingxing
    '''    
    def set_params_from_neptxt(self, file_path="nep.txt"):
        # self.c2_param = None
        # self.c3_param = None
        # self.q_scaler = None
        # self.model_wb = None
        pass
    
    def read_nep_param_from_nep_file(self, nep_in_file:str):
        with open(nep_in_file, 'r') as rf:
            lines = rf.readlines()
        nep_dict = {}
        for line in lines:
            substring = line.split("#")[0].strip()
            if len(substring.split()) > 1:
                key = substring.split()[0]
                value_str = substring.replace(key, "")
                nep_dict[key.lower()] = value_str.strip()
        return nep_dict

    # def to_dict(self):
    #     dicts = {}
    #     dicts["version"] = self.version
    #     dicts["type"] = self.type_num
    #     if self.type_weight is not None:
    #         dicts["type_weight"] = self.type_weight
    #     dicts["model_type"] = self.model_type
    #     dicts["prediction"] = self.prediction
    #     if self.zbl is not None:
    #         dicts["zbl"] = self.zbl
    #     dicts["cutoff"] = self.cutoff
    #     dicts["n_max"] = self.n_max
    #     dicts["basis_size"] = self.basis_size
    #     dicts["l_max"] = self.l_max
    #     dicts["neuron"] = self.neuron
    #     dicts["lambda_1"] = self.lambda_1
    #     dicts["lambda_2"] = self.lambda_2
    #     dicts["lambda_e"] = self.lambda_e
    #     dicts["lambda_f"] = self.lambda_f
    #     dicts["lambda_v"] = self.lambda_v
    #     if self.force_delta is not None:
    #         dicts["force_delta"] = self.force_delta
    #     dicts["batch"] = self.batch
    #     dicts["population"] = self.population
    #     dicts["generation"] = self.generation
    #     return dicts

    def to_nep_in_txt(self):
        content = ""
        content += "version     {}\n".format(self.version)
        content += "type        {}\n".format(self.type)
        # if self.type_weight is not None:
        #     content += "type_weight {}\n".format(self.type_weight)
        content += "model_type  {}\n".format(self.model_type)
        content += "prediction  {}\n".format(self.prediction)
        if self.zbl is not None: #' '.join(map(str, int_list))
            content += "zbl         {}\n".format(self.zbl)
        content += "cutoff      {}\n".format(" ".join(map(str, self.cutoff)))
        content += "n_max       {}\n".format(" ".join(map(str, self.n_max)))
        content += "basis_size  {}\n".format(" ".join(map(str, self.basis_size)))
        content += "l_max       {}\n".format(" ".join(map(str, self.l_max)))
        content += "neuron      {}\n".format(self.neuron[0]) # filter the output layer
        # these are from optimizer SNES
        # content += "lambda_1 {}\n".format(self.lambda_1)
        # content += "lambda_2 {}\n".format(self.lambda_2)
        # content += "lambda_e {}\n".format(self.lambda_e)
        # content += "lambda_f {}\n".format(self.lambda_f)
        # content += "lambda_v {}\n".format(self.lambda_v)
        # if self.force_delta is not None:
        #     content += "force_delta {}\n".format(self.force_delta)
        # content += "batch {}\n".format(self.batch)
        # content += "population {}\n".format(self.population)
        # content += "generation {}\n".format(self.generation)
        return content
    
    def to_nep_txt(self):
        content = ""
        if self.zbl is not None:
            content += "nep4_zbl   {}\n".format(self.type)    #line1
            content += "zbl   {} {}\n".format(self.zbl/2, self.zbl)    #line zbl
        else:
            content += "nep4   {}\n".format(self.type)    #line1
        content += "cutoff {}\n".format(" ".join(map(str, self.cutoff)))    #line2
        content += "n_max  {}\n".format(" ".join(map(str, self.n_max)))    #line3
        content += "basis_size {}\n".format(" ".join(map(str, self.basis_size)))    #line4
        content += "l_max  {}\n".format(" ".join(map(str, self.l_max)))    #line5
        content += "ANN    {} {}\n".format(self.neuron[0], 0)    #line6
        return content

    '''
    description: 
        the format of type in nep.in file is as:
            "type 2 Te Pb # this is a mandatory keyword"
    param {*} self
    param {*} atom_type_list
    return {*}
    author: wuxingxing
    '''    
    def set_atom_type(self, atom_type_list):
        atom_names = []
        for atom in atom_type_list:
            atom_names.append(element_table[atom])
        return str(len(atom_type_list)) + " " + " ".join(atom_names)

    '''
    description: 
        if the value is boath in nep.in file and user input json file, the nep.in value will be used
    param {str} param
    param {dict} json_input
    param {*} default_value
    param {dict} nep_file_dict
    param {str} data_type: int_value 0, float_value 1, int_list 2, float_list 3, string 4
    return {*}
    author: wuxingxing
    '''
    def get_parameter(self, param:str, json_input:dict, default_value, nep_file_dict:dict={}, data_type:int=4):
        if param.lower() in nep_file_dict.keys():
            if data_type == 0:
                return int(nep_file_dict[param])
            elif data_type == 1:
                return float(nep_file_dict[param])
            elif data_type == 2:
                return [int(_) for _ in nep_file_dict[param].split()]
            elif data_type == 3:
                return [float(_) for _ in nep_file_dict[param].split()]
            else:
                return nep_file_dict[param]
            
        if param not in json_input.keys():
            return default_value
        else:
            return json_input[param]


'''
description: 
    get value of param in json_input which is required parameters which need input by user
        if the parameter is not specified in json_input, raise error and print error log to user.
param {str} param
param {dict} json_input
param {str} info 
return {*}
author: wuxingxing
'''
def get_required_parameter(param:str, json_input:dict, detail_infos:str=None):
    if param not in json_input.keys():
        error_info = "Input error! : The {} parameter is missing and must be specified in input json file!".format(param)
        if detail_infos is not None:
            error_info += "\n{}\n".format(detail_infos)
        raise Exception(error_info)
    return json_input[param]

'''
description: 
    get value of param in json_input,
        if the parameter is not specified in json_input, return the default parameter value 'default_value'
param {str} param
param {dict} json_input
param {*} default_value
return {*}
author: wuxingxing
'''
def get_parameter(param:str, json_input:dict, default_value, format_type:str=None):
    res = None
    if param not in json_input.keys():
        res = default_value
    else:
        res = json_input[param]
    
    if format_type is not None:
        if format_type.upper() == "upper".upper():
            res = res.upper()
    return res


def convert_keys_to_lowercase(dictionary:dict):
    if isinstance(dictionary, dict):
        new_dict = {}
        for key, value in dictionary.items():
            new_key = key.lower()
            new_value = convert_keys_to_lowercase(value)  # 递归处理嵌套的字典
            new_dict[new_key] = new_value
        return new_dict
    elif isinstance(dictionary, list):
        return [convert_keys_to_lowercase(item) for item in dictionary]
    else:
        return dictionary


import os

"""
@Description :
parameter:
work_dir: first dir
type_dir: second dir

@Returns     :
@Author       :wuxingxing
"""

class WorkFile(object):
    """
    fixed dir construction :
    work_dir
    --------|train_type_dir
    ---------------------|log_dir
    ---------------------|model_dir
    such as ./work_dir/train_init_drop_10pct_dpnn means: 
            1. the model trained with 10% training data; 
            2. the newwork use dropout strategy;  
            3.the model type is dpnn
    """
    def __init__(self, work_dir) -> None:
        self.work_dir = work_dir
        self.tag_success = "tag.success"
        self.tag_error = "tag.error"

    def check_tag(self, prefix:str):
        if os.path.exists(os.path.join(prefix, self.tag_success)):
            return True
        else:
            return False

class TrainWork(WorkFile):
    def __init__(self, work_dir) -> None:
        super.__init__(work_dir)

    def set_model_init_infos(self, model_num: int=0, model_type: str="DP", init_data_path: str=None):
        self.model_num = model_num
        self.model_type = model_type
        self.init_data_path = init_data_path
    
    def set_train_datas(self, train_data_paths: list[str]):
        self.train_data_paths = train_data_paths

    def get_train_slurm_script(self, model_num_index:int =0):
        self.train_slurm_file = os.path.join("")

        self.log_dir = os.path.join(self.work_dir, "log_dir")
        self.model_dir = os.path.join(self.work_dir, "model_dir")
        self.train_kpu_dir = os.path.join(self.model_dir, "train_0_kpu_dir")
        self.md_kpu_dir = os.path.join(self.model_dir, "md_kpu_dir")

        self.config_yaml_path = os.path.join(self.work_dir, "train.yaml")
        self.md_kpu_config_yaml_path = os.path.join(self.work_dir, "md_kpu.yaml")
        self.md_kpu_script = os.path.join(self.work_dir, "md_kpu_slurm.job")
        self.train_kpu_script = os.path.join(self.work_dir, "train_kpu_slurm.job")
        self.train_script = os.path.join(self.work_dir, "train_slurm.job")
        self.train_success_tag = os.path.join(self.model_dir, "train_success.tag")
        self.md_kpu_success_tag = os.path.join(self.model_dir, "md_kpu_success.tag")
        self.train_kpu_success_tag = os.path.join(self.model_dir, "train_kpu_success.tag")

        self.model_save_path = os.path.join(self.model_dir, "checkpoint.pth.tar")

        # self.tensorbord_dir = os.path.join(self.work_dir, "tensorbord_dir")
        if os.path.exists(self.work_dir) is False:
            os.makedirs(self.work_dir)
        if os.path.exists(self.log_dir) is False:
            os.makedirs(self.log_dir)
        if os.path.exists(self.model_dir) is False:
            os.makedirs(self.model_dir)
        
    
    """
    @Description :
    setting training data files path
    @Returns     :
    @Author       :wuxingxing
    """
    def set_torch_data_path(self, data_path = None):
        self.data_path = data_path
        self.train_data_path = []
        self.valid_data_path = []
        for i in range(len(self.data_path)):
            self.train_data_path.append(os.path.join(self.data_path[i], 'train_data/final_train'))
            self.valid_data_path.append(os.path.join(self.data_path[i], 'train_data/final_test'))
        
        print("work torch data paths are {}".format(self.data_path))

    def set_torch_data_sacle_infos(self, davg_dstd_dir):
        self.davg_dstd_dir = davg_dstd_dir
        
    def print_dir_tree(self):
        # latter would change by os.walk to prient the workdir tree
        print("the work dir is as listed:")
        print("work_dir:{}".format(self.work_dir))
        print("second_dir:{}\t{}\t{}\t".format(self.log_dir, self.model_dir, self.tensorbord_dir))
        print("set_torch_data_path:{}".format(self.data_path))

class WorkValidDir(object):
    """
    fixed dir construction :
    work_dir
    --------|train_type_dir
    ---------------------|log_dir
    ---------------------|model_dir
    ---------------------|tensorbord_record_dir
    such as ./work_dir/train_init_drop_10pct_dpnn means: 
            1. the model trained with 10% training data; 
            2. the newwork use dropout strategy;  
            3.the model type is dpnn
    """
    def __init__(self, work_path, model_path, p_path) -> None:
        self.work_dir = work_path
        self.model_dir = model_path
        self.p_dir = p_path
        self.data_path = None
        self.train_data_path = None
        self.valid_data_path = None

        if os.path.exists(self.work_dir) is False:
            os.makedirs(self.work_dir)

    """
    @Description :
    setting training data files path
    @Returns     :
    @Author       :wuxingxing
    """
    def set_torch_data_path(self, data_path = None):
        self.data_path = data_path
        self.train_data_path = []
        self.valid_data_path = []
        for i in range(len(self.data_path)):
            self.train_data_path.append(os.path.join(self.data_path[i], 'train_data/final_train'))
            self.valid_data_path.append(os.path.join(self.data_path[i], 'train_data/final_test'))
        
        print("work torch data paths are {}".format(self.data_path))

    def set_torch_data_sacle_infos(self, davg_dstd_dir):
        self.davg_dstd_dir = davg_dstd_dir
        

class WorkMDDir(object):
    def __init__(self, root_dir) -> None:
        self.work_dir = root_dir
        if os.path.exists(self.work_dir) is False:
            os.makedirs(self.work_dir)

        self.log_dir = os.path.join(self.work_dir, "log_dir")
        self.md_dir = os.path.join(self.work_dir, "md_dir")
        self.md_success_tag = os.path.join(self.md_dir, "md_success.tag")
        self.md_slurm_path = os.path.join(self.md_dir, "md_slurm.job")
        
        self.md_dpkf_dir = os.path.join(self.work_dir, "md_dpkf_dir") #the dir of pwmat+dpkf-movemenet to dpkf train for kpu analyse
        self.md_traj_dir = os.path.join(self.work_dir, "md_traj_dir")   #seperate movement to each atom_i.config and save to this path
        
        if os.path.exists(self.log_dir) is False:
            os.makedirs(self.log_dir)
        if os.path.exists(self.md_dir) is False:
            os.makedirs(self.md_dir)
        if os.path.exists(self.md_dpkf_dir) is False:
            os.makedirs(self.md_dpkf_dir)
        if os.path.exists(self.md_traj_dir) is False:
            os.makedirs(self.md_traj_dir)
        
        self.set_md_files()

    def set_md_files(self):
        self.model_path = os.path.join(self.md_dir, "checkpoint.pth.tar")
        self.p_path = os.path.join(self.md_dir, "P.pt")
        self.atom_config_path = os.path.join(self.md_dir, "atom.config")
        self.fread_dfeat_path = os.path.join(self.md_dir, "fread_dfeat")
        self.feat_info_path = os.path.join(self.fread_dfeat_path, "feat.info")
        self.vdw_fitB_ntype_path = os.path.join(self.fread_dfeat_path, "vdw_fitB.ntype")
        self.md_input_path = os.path.join(self.md_dir, "md.input")
        self.gen_feat_path = os.path.join(self.md_dpkf_dir, "gen_feat.job")
        self.gen_feat_success_tag = os.path.join(self.md_dpkf_dir, "gen_feat_success.tag")

    def print_dir_tree(self):
        print("the work dir is as listed:")
        print("work_dir:{}".format(self.work_dir))
        print("second_dir:{}\t{}\t".format(self.log_dir, self.md_dir))

class WorkLabDir(object):
    def __init__(self, root_dir) -> None:
        self.work_dir = root_dir
        self.ab_dir = os.path.join(self.work_dir, "ab_dir")
        self.lab_dpkf_dir = os.path.join(self.work_dir, "lab_dpkf_dir")
        self.scf_slurm_path = os.path.join(self.ab_dir, "scf_slurm.job")
        self.scf_success_tag = os.path.join(self.ab_dir, "scf_success.tag")

        self.gen_feat_slurm_path = os.path.join(self.lab_dpkf_dir, "gen_feat_slurm.job")
        self.gen_feat_success_tag = os.path.join(self.lab_dpkf_dir, "gen_feat_success.tag")

        if os.path.exists(self.work_dir) is False:
            os.makedirs(self.work_dir)
        if os.path.exists(self.ab_dir) is False:
            os.makedirs(self.ab_dir)
        if os.path.exists(self.lab_dpkf_dir) is False:
            os.makedirs(self.lab_dpkf_dir)

if __name__=="__main__":
    print()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pwact", 
    version="0.4.6",
    author="LonxunQuantum",
    author_email="lonxun@pwmat.com",
    description="PWACT is an open-source automated active learning platform based on MatPL for efficient data sampling.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LonxunQuantum/PWact",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    entry_points={
        'console_scripts': [
            'pwact = pwact.main:main'
        ]
    }
)




import subprocess
import os

def get_job_ids():
    # 使用 squeue 获取当前用户的所有作业ID列表
    user = os.getlogin()  # 获取当前登录的用户名
    result = subprocess.run(['squeue', '-u', user, '-o', '%i', '-h'], capture_output=True, text=True)
    job_ids = result.stdout.strip().split()
    return job_ids

def get_slurm_script_path(job_id):
    # 使用 scontrol 显示作业详细信息，并从中提取 slurm 脚本路径
    result = subprocess.run(['scontrol', 'show', 'job', str(job_id)], capture_output=True, text=True)
    for line in result.stdout.splitlines():
        if "Command=" in line:
            script_path = line.split('=')[1]
            return script_path
    return None

def get_jobs(work_dir:str):
    job_ids = get_job_ids()
    jobs = []
    for job_id in job_ids:
        script_path = get_slurm_script_path(job_id)
        # if script_path:
        #     print(f"Job ID: {job_id}, Slurm Script Path: {script_path}")
        # else:
        #     print(f"Job ID: {job_id}, Slurm Script Path: Not Found")
        if work_dir in script_path:
            jobs.append(job_id)
    return jobs
        

# if __name__ == "__main__":
#     get_jobs("run_iter/iter.0000/temp_run_iter_work/label/scf")


import os
import numpy as np

# import numpy as np
import matplotlib.pyplot as plt

mark_list = ["o", "^", "v", "X", "|", "*", "v", "+", '*', ' ']
color_list = ["#006400", "#FF8C00", "#B22222", "#FF8C00" ,"#ff0099" ,\
              "#000000" , "#BDB76B", "#999900" ,"#009999" ,"#000099" ,"#990099", "#ff9900" ]
linestyle_list = ["--", "--", "--", "-", "-", "-"]

def draw_hists_2(data, save_path, min, max, legend_label):
    bins = np.linspace(0, max, 100)  # 根据需要调整bin的数量和范围
    alpha = 0.5  # 透明度设置
    plt.figure(figsize=(10, 6))  # 设置图形的大小

    for i, _data in enumerate(data):
        plt.hist(_data, bins=bins, alpha=alpha, label=legend_label[i], density=True)

    # 设置坐标轴范围
    plt.xlim(min, max)
    
    # 添加标题、标签、图例
    plt.title('Distribution of model deviation', fontsize=18, fontweight='bold', family='serif')
    plt.xlabel('Model Deviation', fontsize=14, fontweight='bold', family='serif')
    plt.ylabel('Frequency', fontsize=14, fontweight='bold', family='serif')
    
    # 设置图例
    plt.legend(loc='upper right', fontsize=15, title_fontsize='large')

    # 调整坐标轴的字体大小
    plt.tick_params(axis='both', labelsize=15)

    # 添加网格
    plt.grid(True, linestyle='--', alpha=0.7)

    # 保存图像
    plt.savefig(save_path, dpi=300, bbox_inches='tight')
    plt.close()

def read_model_devi(file_path, low=None, high=None):
    devi_files = []
    for path, dirList, fileList in os.walk(file_path):
        for _ in fileList:
            if "model_devi.out" in _:
                devi_files.append(os.path.join(path, _))
    data = []
    for file in devi_files:
        _data = np.loadtxt(file, skiprows=0)
        if _data.shape[0] > 1:
            data.extend(list(_data[:, 3])) # force
    
    right = 0
    mid = 0
    error = 0
    if low is not None:
        for d in data:
            if d <= low:
                right += 1
            elif d > low and d <= high:
                mid += 1
            else:
                error += 1
    return data, devi_files, right, mid, error
    
def draw_hist_list(file_path:list[str], legend_label, save_path:str, low=None, high=None) :
    data = []
    min = None
    max = None
    abs_mean = []
    abs_max = []
    mid = 0
    error = 0
    right = 0
    for file in file_path:
        _data, devi_files, _right, _mid, _error = read_model_devi(file, low, high)
        right += _right
        mid += _mid
        error += _error
        _min = np.min(_data)
        _max = np.max(_data)
        min = _min if (min is None or _min < min) else min
        max = _max if (max is None or _max > max) else max
        data.append(_data)
        abs_mean.append(np.mean(_data))
        abs_max.append(np.max(_data))
    
    for i in range(0, len(legend_label)):
        if low is not None:
            legend_label[i] = "{}\nmean {} max {} right {} candidate {} error {}".format(legend_label[i], round(abs_mean[i], 3), round(abs_max[i], 3), right, mid, error)
        else:
            legend_label[i] = "{}\nmean {} max {}".format(legend_label[i], round(abs_mean[i], 5), round(abs_max[i], 5))
    # print(legend_label)
    new_data = []
    for id, d in enumerate(data):
        tmp = []
        for _d in d:
            if _d > 1.0:
                tmp.append(1.0)
            else:
                tmp.append(_d)
        new_data.append(tmp)
    max = 1.0 if max > 1.0 else max
    draw_hists_2(new_data, save_path, min, max, legend_label)




if __name__ == "__main__":
    legend_label = ["iter.0006/md/md.000.sys.000"]
    save_dir = "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.002"
    draw_hist_list([
        "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.002"
    ], legend_label, save_path=os.path.join(save_dir, "hist_mode_deiv_md.000.sys.002.png"))

    legend_label = ["iter.0006/md/md.000.sys.001"]
    save_dir = "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.000"
    draw_hist_list([
        "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.000"
    ], legend_label, save_path=os.path.join(save_dir, "hist_mode_deiv_md.000.sys.000.png"))

    legend_label = ["iter.0006/md/md.000.sys.002"]
    save_dir = "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.001"
    draw_hist_list([
        "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.001"
    ], legend_label, save_path=os.path.join(save_dir, "hist_mode_deiv_md.000.sys.001.png"))

    legend_label = ["iter.0006/md/md.000.sys.003"]
    save_dir = "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.003"
    draw_hist_list([
        "/data/home/wuxingxing/datas/al_dir/djp/iter.0006/explore/md/md.000.sys.003"
    ], legend_label, save_path=os.path.join(save_dir, "hist_mode_deiv_md.000.sys.003.png"), low=0.1, high=0.2)



'''
description: 
    Active learning of top-level directory structure
return {*}
author: wuxingxing
'''
class AL_WORK:
    init_bulk = "init_bulk"
    init_surface = "init_surface"
    run_iter = "run"

class AL_STRUCTURE:
    train = "00.train"
    explore = "01.explore"
    labeling = "02.label"
    pertub = "pertub"
    aimd = "aimd"
    collection = "collection"
    init_data = "init_data_npy"

'''
description: 
    slurm job log file name
return {*}
author: wuxingxing
'''
class SLURM_OUT:
    dft_out = "dft.log"
    train_out = "train.log"
    md_out = "md.log"
    kpu_out = "kpu.log"

class SLURM_JOB_TYPE:
    cp2k_relax = "cp2k/relax"
    cp2k_scf = "cp2k/scf"
    cp2k_aimd = "cp2k/aimd"
    pwmat_relax = "pwmat/relax"
    pwmat_scf = "pwmat/scf"
    pwmat_aimd = "pwmat/aimd"
    vasp_relax = "vasp/relax"
    vasp_scf = "vasp/scf"
    vasp_aimd = "vasp/aimd"
    lammps = "lammps"
    direct = "direct"

'''
description: 
    training commands
return {*}
author: wuxingxing
'''
class MODEL_CMD:
    train = "train"
    gen_feat = "gen_feat"
    test = "test"
    script = "script"
    compress = "compress"
    kpu = "kpu"
    pwdata =  "pwdata"

'''
description: 
    model training parametors
return {*}
author: wuxingxing
'''
class TRAIN_INPUT_PARAM:
    save_p_matrix = "save_p_matrix" # for kpu
    raw_files = "raw_files"
    valid_data = "valid_data"
    train_data = "train_data"
    format = "format"
    test_mvm_files = "test_movement_file"
    reserve_feature = "reserve_feature" #False
    reserve_work_dir = "reserve_work_dir" #False
    valid_shuffle = "valid_shuffle" #True
    train_valid_ratio = "train_valid_ratio" #0.8
    sample_nums ="sample_nums" # for kpu or test nums
    seed = "seed" #2023
    recover_train = "recover_train" #true
    type_embedding = "type_embedding"
    model_type = "model_type"
    atom_type = "atom_type"
    model_load_file = "model_load_file"
    test_dir_name = "test_dir_name"
    work_dir = "work_dir"
    optimizer = "optimizer"
    reset_epoch = "reset_epoch"
    #epoch

'''
description: 
    lammps md commands
return {*}
author: wuxingxing
'''
class LAMMPS_CMD:
    lmp_mpi = "lmp_mpi"
    lmp_mpi_gpu = "lmp_mpi_gpu"

'''
description: 
    force filed types
return {*}
author: wuxingxing
'''
class FORCEFILED:
    fortran_lmps = 1 # use cpu dp model
    libtorch_lmps = 2 # default, use jit model or nep cpu/gpu model
    main_md = 3 #

'''
description: 
    pwdata format
return {*}
author: wuxingxing
'''
class PWDATA:
    pwmat_config = "pwmat/config"
    vasp_poscar = "vasp/poscar"
    lammps_dump = "lammps/dump"
    lammps_lmp = "lammps/lmp"
    pwmat_movement = "pwmat/movement"
    vasp_outcar = "vasp/outcar"
    extxyz = "extxyz"
    vasp_xml = "vasp/xml"
    cp2k_md = 'cp2k/md'
    cp2k_scf = 'cp2k/scf'
    pwmlff_npy = "pwmlff/npy"

'''
description: 
    DFT apps command
return {*}
author: wuxingxing
'''
class DFT_STYLE:
    vasp = "vasp"
    pwmat = "pwmat"
    cp2k = "cp2k"
    lammps = "lammps"
    bigmodel="bigmodel"

    '''
    description: 
        is_cp2k_coord: if ture, for cp2k, return 'cp2k/scf', it is used to extract config from cp2k log file
    return {*}
    author: wuxingxing
    '''
    @staticmethod
    def get_pwdata_format(dft_style:str, is_cp2k_coord:bool=False):
        if dft_style.lower() == DFT_STYLE.pwmat.lower() : # atom.config
            return PWDATA.pwmat_config
        if dft_style.lower() == DFT_STYLE.vasp.lower(): 
            return PWDATA.vasp_poscar
        if dft_style.lower() == DFT_STYLE.cp2k.lower(): 
            if is_cp2k_coord:
                return PWDATA.cp2k_scf
            else:
                return PWDATA.vasp_poscar
        if dft_style.lower() == DFT_STYLE.bigmodel.lower():
            return PWDATA.extxyz

    @staticmethod
    def get_normal_config(dft_style:str): # the input config file name of pwmat vasp and cp2k
        if dft_style == DFT_STYLE.pwmat: # atom.config
            return PWMAT.atom_config
        elif dft_style == DFT_STYLE.vasp: # poscar
            return VASP.poscar
        elif dft_style == DFT_STYLE.cp2k: # coord.xyz, for cp2k, the position of atoms writed in coord.xyz file, the cell is in inp file
            return CP2K.coord_xyz
    
    @staticmethod
    def get_postfix(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return ".config"
        elif dft_style == DFT_STYLE.vasp:
            return ".poscar"
        elif dft_style == DFT_STYLE.cp2k:
            return ".poscar"

    @staticmethod
    def get_format_by_postfix(file_name:str):
        if "config" in file_name.lower()\
            or "pwmat" in file_name.lower():
            return PWDATA.pwmat_config
        elif "movement" in file_name.lower():
            return PWDATA.pwmat_movement
        elif "out.mlmd" in file_name.lower():
            return PWDATA.pwmat_movement
            
        elif "poscar" in file_name.lower() or\
            "contcor" in file_name.lower() or\
                "vasp" in file_name.lower():
            return PWDATA.vasp_poscar

        elif "outcar" in file_name.lower():
            return PWDATA.vasp_outcar

        elif "inp" in file_name.lower() or\
            "xyz" in file_name.lower() or\
                "cp2k" in file_name.lower():
            return PWDATA.cp2k_scf

        elif CP2K.final_config.lower() in file_name.lower():
            return PWDATA.cp2k_scf
        
    '''
    description: 
        for pwmat is final.config
        for vasp is  CONTCAR
    return {*}
    author: wuxingxing
    '''    
    @staticmethod
    def get_relaxed_original_name(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return PWMAT.final_config
        elif dft_style == DFT_STYLE.vasp:
            return VASP.final_config
        elif dft_style == DFT_STYLE.cp2k:
            return CP2K.final_config


    @staticmethod
    def get_scf_config(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return PWMAT.out_mlmd
        elif dft_style == DFT_STYLE.vasp:
            return VASP.outcar
        elif dft_style == DFT_STYLE.cp2k:
            return CP2K.final_config

    '''
    description: 
        scf files need reserved
            for vasp is outcar poscar and incar file
            for pwmat is atom.config etot.input report and out.mlmd file
                the scf of pwmat will produce a file named final.config, does not reserve
    param {str} dft_style
    return {*}
    author: wuxingxing
    '''
    @staticmethod
    def get_scf_reserve_list(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            scf_list = PWMAT.scf_reserve_list
        elif dft_style == DFT_STYLE.vasp:
            scf_list = VASP.scf_reserve_list
        elif dft_style == DFT_STYLE.cp2k:
            scf_list = CP2K.scf_reserve_list

        scf_list = [_.lower() for _ in scf_list]
        return scf_list
    
    @staticmethod
    def get_aimd_config(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return PWMAT.MOVEMENT
        elif dft_style == DFT_STYLE.vasp:
            return VASP.outcar
        elif dft_style == DFT_STYLE.cp2k:# for cp2k, convert the output content to poscar format
            return CP2K.final_config

    @staticmethod
    def get_aimd_config_format(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return PWDATA.pwmat_movement
        elif dft_style == DFT_STYLE.vasp:
            return PWDATA.vasp_outcar
        elif dft_style == DFT_STYLE.cp2k:# for cp2k, convert the output content to poscar format
            return PWDATA.cp2k_md

    @staticmethod
    def get_pertub_config(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return "pertub.config"
        elif dft_style == DFT_STYLE.vasp:
            return "pertub.poscar"
        elif dft_style == DFT_STYLE.cp2k:# for cp2k, convert the output content to poscar format
            return "pertub.poscar"

    @staticmethod
    def get_super_cell_config(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return "super_cell.config"
        elif dft_style == DFT_STYLE.vasp:
            return "super_cell.poscar"
        elif dft_style == DFT_STYLE.cp2k:# for cp2k, convert the output content to poscar format
            return "super_cell.poscar"

    
    @staticmethod
    def get_scale_config(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return "scale.config"
        elif dft_style == DFT_STYLE.vasp:
            return "scale.poscar"
        elif dft_style == DFT_STYLE.cp2k:# for cp2k, convert the output content to poscar format
            return "scale.poscar"

    @staticmethod
    def get_relaxed_config(dft_style:str):
        if dft_style == DFT_STYLE.pwmat:
            return "relaxed.config"
        elif dft_style == DFT_STYLE.vasp:
            return "relaxed.poscar"
        elif dft_style == DFT_STYLE.cp2k: # for cp2k, convert the output content to poscar format
            return "relaxed.poscar"
    
class DFT_TYPE:
    relax = "relax"
    aimd = "aimd"
    scf = "scf"

'''
description: 
    temp files name
return {*}
author: wuxingxing
'''
class TEMP_STRUCTURE:
    tmp_init_bulk_dir = "temp_init_bulk_work"
    tmp_run_iter_dir = "temp_run_iter_work"
    tmp_prefix = "temp"
    
class INIT_BULK:
    relax = "relax"
    relax_job ="relax.job"
    relax_tag = "tag.relax.success"
    relax_tag_failed = "tag.relax.failed"
    
    init_config = "init_config"
    init = "init"
    super_cell_scale = "super_cell_scale"
    tag_super_cell = "tag.success.super_cell_scale"
    super_cell = "super_cell"
    scale = "scale"
    
    pertub = "pertub"
    tag_pertub = "tag.success.pertub"
    
    aimd = "aimd"
    aimd_job = "aimd.job"
    aimd_tag = "tag.aimd.success"
    aimd_tag_failed ="tag.aimd.failed"
    
    scf = "scf"
    scf_job = "scf.job"
    scf_tag = "tag.scf.success"
    scf_tag_failed ="tag.scf.failed"

    bigmodel="bigmodel"
    bigmodel_job = "bigmodel.job"
    bigmodel_tag = "tag.bigmodel.success"
    bigmodel_tag_failed ="tag.bigmodel.failed"
    bigmodel_traj = "traj.xyz"
    
    direct="direct"
    direct_job = "direct.job"
    direct_tag = "tag.direct.success"
    direct_tag_failed ="tag.direct.failed"
    candidate_xyz="candidate.xyz"
    candidate_idx="candidate.json"
    direct_traj = "select.xyz"

    collection = "collection"
    npy_format_save_dir = "PWdata"
    npy_format_name = "datapath.txt"

    def get_save_format(save_format:str):
        if save_format == PWDATA.pwmlff_npy:
            return INIT_BULK.npy_format_save_dir
        elif save_format==PWDATA.extxyz:
            return "train.xyz"
        else:
            raise Exception("Error! The data_format only support ")

class MODEL_TYPE:
    dp = "DP"
    nep = "NEP"

class TRAIN_FILE_STRUCTUR:
    work_dir = "work_dir"
    feature_dir = "feature"
    feature_json = "feature.json"
    feature_job = "feature.job"
    feature_tag = "tag.feature.success"
    feature_tag_failed = "tag.feature.failed"
    train_json = "train.json"
    train_job = "train.job"
    train_tag = "tag.train.success"
    train_tag_failed = "tag.train.failed"
    
    kpu_json = "kpu.json"
    kpu_tag = "tag.kpu.success"
    kpu_tag_failed = "tag.kpu.failed"
    kpu_job = "kpu.job"
    kpu = "kpu"
    kpu_file = "kpu_info.csv"
    base_kpu = "base_kpu"
    
    movement = "MOVEMENT"
    model_record = "model_record"
    # dp model
    dp_model_name ="dp_model.ckpt"
    compree_dp_name = "cmp_dp_model.ckpt"
    # cmp_tracing_dp_name = "torch_script_module.pt"
    script_dp_name = "torch_script_module.pt"
    script_dp_name_cpu = "jit_dp_cpu.pt"
    script_dp_name_gpu = "jit_dp_gpu.pt"
    fortran_dp = "forcefield"
    fortran_dp_name = "forcefield.ff"

    # nep model
    nep_model_name ="nep_model.ckpt"
    nep_model_lmps = "nep5.txt"

class EXPLORE_FILE_STRUCTURE:
    kpu= "kpu"
    md = "md"
    select = "select"
    direct = "direct"
    md_tag = "tag.md.success"
    md_tag_faild = "tag.md.error"
    md_job = "md.job"
    direct_tag = "tag.direct.success"
    direct_tag_faild = "tag.direct.error"
    direct_job = "direct.job"
    # selected image info file names
    candidate = "candidate.csv"
    candidate_xyz="candidate.xyz"
    select_idx = "select_idx.dat"
    select_xyz = "select.xyz"
    # candidate_random = "candidate_random.csv"
    candidate_delete = "candidate_delete.csv"
    failed = "fail.csv"
    accurate = "accurate.csv"
    select_summary = "select_summary.txt"

    traj ="traj"
    
    # for committee and kpu method 
    model_devi = "model_devi.out"
    md_traj_error_record="error_traj.log"
    kpu_model_devi = "kpu_model_devi.out"
    devi_columns = ["devi_force", "config_index", "file_path"]

    iter_select_file = "iter_select.txt"

    @staticmethod
    def get_devi_name(data_type:str):
        if data_type == UNCERTAINTY.committee:
            devi_name = EXPLORE_FILE_STRUCTURE.model_devi
        elif data_type == UNCERTAINTY.kpu:
            devi_name = EXPLORE_FILE_STRUCTURE.kpu_model_devi
        else:
            raise Exception("get_devi_name error, the data_type {} not relized".format(data_type))
        return devi_name


class LABEL_FILE_STRUCTURE:
    bigmodel="bigmodel"
    scf = "scf"
    result = "result"
    scf_tag = "tag.scf.success"
    scf_tag_failed = "tag.scf.failed"
    scf_job = "scf.job"
    bigmodel_job="bigmodel.job"
    bigmodel_tag = "tag.bigmodel.success"
    bigmodel_tag_failed = "tag.bigmodel.failed"
    train_xyz = "train.xyz"


class LAMMPS:
    input_lammps="in.lammps"
    poscar = "POSCAR"
    lammps_sys_config = "lmp.config"
    traj_postfix = ".lammpstrj"
    log_lammps = "log.lammps"
    traj_format = "dump"
    lmp_format = "lmp"
    atom_type_file = "atom_type.txt"

class ENSEMBLE:
    npt_tri = "npt_tri",
    nvt = "nvt"
    npt = "npt"
    nve = "nve"

class PWMAT:
    in_skf = "in_skf"
    pwmat_out = "PWMAT.out"
    config_postfix = ".config"
    atom_config = "atom.config"
    etot_input = "etot.input"
    out_mlmd = "OUT.MLMD"
    mvm = "mvm"
    relax="RELAX"
    scf = "SCF"
    md = "MD"
    MOVEMENT="MOVEMENT"
    MOVEMENT_low = "movement"
    kspacing_default = 0.5
    scf_reserve_list = ["REPORT", "etot.input","OUT.MLMD", ".config", "MOVEMENT"]
    final_config = "final.config"#relaxed result

class VASP:
    potcar = "POTCAR"
    incar = "INCAR"
    poscar = "POSCAR"
    final_config = "CONTCAR"#relaxed result
    outcar = "OUTCAR"
    scf_reserve_list = ["OUTCAR","POSCAR", "INCAR"]

class CP2K:
    cp2k_inp  = "cp2k.inp"
    coord_xyz = "coord.xyz"
    cell_txt = "cell.txt"
    traj_xyz = "traj.xyz"
    final_config = "dft.log" # relaxed result , the cp2k output should be extraced from log
    scf_reserve_list = ["cp2k.inp", "dft.log", "coord.xyz"] # scf reserve file list

class UNCERTAINTY:
    kpu = "KPU"
    committee="COMMITTEE"

ELEMENTTABLE={'H': 1,
    'He': 2,  'Li': 3,  'Be': 4,  'B': 5,   'C': 6,   'N': 7,   'O': 8,   'F': 9,   'Ne': 10,  'Na': 11,
    'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15,  'S': 16,  'Cl': 17, 'Ar': 18, 'K': 19,  'Ca': 20,  'Sc': 21,
    'Ti': 22, 'V': 23,  'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30,  'Ga': 31,
    'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39,  'Zr': 40,  'Nb': 41, 
    'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50,  'Sb': 51, 
    'Te': 52, 'I': 53,  'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60,  'Pm': 61,
    'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70,  'Lu': 71, 
    'Hf': 72, 'Ta': 73, 'W': 74,  'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80,  'Tl': 81, 
    'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90,  'Pa': 91, 
    'U': 92,  'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100, 'Md': 101,
    'No': 102,'Lr': 103,'Rf': 104,'Db': 105,'Sg': 106,'Bh': 107,'Hs': 108,'Mt': 109,'Ds': 110,'Rg': 111,
    'Uub': 112
    }

ELEMENTTABLE_2 = {1: 'H', 
    2: 'He',     3: 'Li',   4: 'Be',   5: 'B',    6: 'C',    7: 'N',   8: 'O',     9: 'F',   10: 'Ne',  11: 'Na', 
    12: 'Mg',   13: 'Al',  14: 'Si',  15: 'P',   16: 'S',   17: 'Cl',  18: 'Ar',  19: 'K',   20: 'Ca',  21: 'Sc', 
    22: 'Ti',   23: 'V',   24: 'Cr',  25: 'Mn',  26: 'Fe',  27: 'Co',  28: 'Ni',  29: 'Cu',  30: 'Zn',  31: 'Ga', 
    32: 'Ge',   33: 'As',  34: 'Se',  35: 'Br',  36: 'Kr',  37: 'Rb',  38: 'Sr',  39: 'Y',   40: 'Zr',  41: 'Nb', 
    42: 'Mo',   43: 'Tc',  44: 'Ru',  45: 'Rh',  46: 'Pd',  47: 'Ag',  48: 'Cd',  49: 'In',  50: 'Sn',  51: 'Sb', 
    52: 'Te',   53: 'I',   54: 'Xe',  55: 'Cs',  56: 'Ba',  57: 'La',  58: 'Ce',  59: 'Pr',  60: 'Nd',  61: 'Pm', 
    62: 'Sm',   63: 'Eu',  64: 'Gd',  65: 'Tb',  66: 'Dy',  67: 'Ho',  68: 'Er',  69: 'Tm',  70: 'Yb',  71: 'Lu', 
    72: 'Hf',   73: 'Ta',  74:  'W',  75: 'Re',  76: 'Os',  77: 'Ir',  78: 'Pt',  79: 'Au',  80: 'Hg',  81: 'Tl', 
    82: 'Pb',   83: 'Bi',  84: 'Po',  85: 'At',  86: 'Rn',  87: 'Fr',  88: 'Ra',  89: 'Ac',  90: 'Th',  91: 'Pa', 
    92: 'U',    93: 'Np',  94: 'Pu',  95: 'Am',  96: 'Cm',  97: 'Bk',  98: 'Cf',  99: 'Es', 100: 'Fm', 101: 'Md', 
    102: 'No', 103: 'Lr', 104: 'Rf', 105: 'Db', 106: 'Sg', 107: 'Bh', 108: 'Hs', 109: 'Mt', 110: 'Ds', 111: 'Rg', 
    112: 'Uub'
    }

ELEMENTMASSTABLE={  1:1.007,2:4.002,3:6.941,4:9.012,5:10.811,6:12.011,
                            7:14.007,8:15.999,9:18.998,10:20.18,11:22.99,12:24.305,
                            13:26.982,14:28.086,15:30.974,16:32.065,17:35.453,
                            18:39.948,19:39.098,20:40.078,21:44.956,22:47.867,
                            23:50.942,24:51.996,25:54.938,26:55.845,27:58.933,
                            28:58.693,29:63.546,30:65.38,31:69.723,32:72.64,33:74.922,
                            34:78.96,35:79.904,36:83.798,37:85.468,38:87.62,39:88.906,
                            40:91.224,41:92.906,42:95.96,43:98,44:101.07,45:102.906,46:106.42,
                            47:107.868,48:112.411,49:114.818,50:118.71,51:121.76,52:127.6,
                            53:126.904,54:131.293,55:132.905,56:137.327,57:138.905,58:140.116,
                            59:140.908,60:144.242,61:145,62:150.36,63:151.964,64:157.25,65:158.925,
                            66:162.5,67:164.93,68:167.259,69:168.934,70:173.054,71:174.967,72:178.49,
                            73:180.948,74:183.84,75:186.207,76:190.23,77:192.217,78:195.084,
                            79:196.967,80:200.59,81:204.383,82:207.2,83:208.98,84:210,85:210,86:222,
                            87:223,88:226,89:227,90:232.038,91:231.036,92:238.029,93:237,94:244,
                            95:243,96:247,97:247,98:251,99:252,100:257,101:258,102:259,103:262,104:261,105:262,106:266}

def get_atomic_number_from_name(atomic_names:list[str]):
    res = []
    for name in atomic_names:
        res.append(ELEMENTTABLE[name])
    return res

def get_atomic_name_from_number(atomic_number:list[int]):
    res = []
    for number in atomic_number:
        res.append(ELEMENTTABLE_2[number])
    return res

def get_atomic_name_from_str(atom_strs):
    try:
        return [int(_) for _ in atom_strs]
    except ValueError:
        return get_atomic_number_from_name(atom_strs)


# print(get_atomic_name_from_str([8, 72]))

# print(get_atomic_name_from_str(["8", "72"]))

# print(get_atomic_name_from_str(["O", "Hf"]))



"""
    dir of explore:
    iter.0000/explore/md:
    --------------------/*-md.job file
    --------------------/*-md.tag.success file
    --------------------/md.000.sys.000/ dir
    -----------------------------------/md.000.sys.000.t.000.p.000 or md.000.sys.000.t.000 dir
    -------------------------------------------------------------/md files: lmp.config, in.lammps, forcefile files, model_devi file
    -------------------------------------------------------------/trajs
    iter.0000/explore/select: this is kpu work dir
    -------------------/md.000.sys.000/
    ----------------------------------/md.000.sys.000.t.000(p.000)
    -------------------------------------------------------------/kpu.job kpu.json dp_model.ckpt
    -------------------------------------------------------------kpu dir
    --------------------------------------------------------------------------/kpu_info.csv
    after select
    -------------------/summary.txt
    -------------------/accurate.txt candidate.txt failed.txt candidate_del.txt
    
    the content of candidate.txt is:
"""
import os
import glob
import numpy as np
import pandas as pd

from pwact.active_learning.slurm.slurm import SlurmJob, Mission
from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.iter_input import InputParam
from pwact.active_learning.explore.select_image import select_image
from pwact.utils.format_input_output import make_train_name, get_iter_from_iter_name, get_sub_md_sys_template_name,\
    get_md_sys_template_name
from pwact.utils.constant import AL_STRUCTURE, TRAIN_FILE_STRUCTUR, MODEL_CMD, \
    EXPLORE_FILE_STRUCTURE, LAMMPS, SLURM_OUT, TEMP_STRUCTURE, UNCERTAINTY

from pwact.utils.file_operation import write_to_file, link_file, search_files, del_file_list_by_patten
from pwact.utils.slurm_script import get_slurm_job_run_info, split_job_for_group, set_slurm_script_content
from pwact.active_learning.explore.select_image import select_image
'''
description: model training method:
1. go to train data path
2. set training command
3. training
4. post process
param {*} itername
param {*} work_type
return {*}
'''

class ModelKPU(object):
    def __init__(self, itername:str, resource: Resource, input_param:InputParam):
        self.itername = itername
        self.resource = resource
        self.input_param = input_param
        # train work dir
        self.train_dir = os.path.join(self.input_param.root_dir, itername, AL_STRUCTURE.train)

        # md work dir
        self.iter = get_iter_from_iter_name(self.itername)
        self.md_job = self.input_param.explore.md_job_list[self.iter]
        # md work dir
        self.explore_dir = os.path.join(self.input_param.root_dir, itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.explore)
        self.md_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.md)
        self.select_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.select)
        self.kpu_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.kpu) # for kpu calculate
      
        self.real_explore_dir = os.path.join(self.input_param.root_dir, itername, AL_STRUCTURE.explore)
        self.real_md_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.md)
        self.real_select_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.select)
        self.real_kpu_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.kpu) # for kpu calculate
 
        
    def make_kpu_work(self):
        kpu_list = []
        
        model_i = make_train_name(0) # train/train.000
        model_i_dir = os.path.join(self.train_dir, model_i)
        source_model_path = os.path.join(model_i_dir, TRAIN_FILE_STRUCTUR.model_record, TRAIN_FILE_STRUCTUR.dp_model_name)

        md_sys_dir_list = search_files(self.md_dir, get_md_sys_template_name())
        for md_sys_dir in md_sys_dir_list:
            sub_md_sys_dir_list =search_files(md_sys_dir, get_md_sys_template_name())
            for sub_md_sys in sub_md_sys_dir_list: #kpu_work_dir
                # set kpu work
                # link kpu calculate model
                kpu_dir = os.path.join(self.kpu_dir, os.path.basename(md_sys_dir), os.path.basename(sub_md_sys))
                if not os.path.exists(kpu_dir):
                    os.makedirs(kpu_dir)
                
                link_file(source_model_path, os.path.join(kpu_dir, TRAIN_FILE_STRUCTUR.dp_model_name))
                link_file(os.path.join(sub_md_sys, LAMMPS.atom_type_file), os.path.join(kpu_dir, LAMMPS.atom_type_file))
                link_file(os.path.join(sub_md_sys, EXPLORE_FILE_STRUCTURE.traj), os.path.join(kpu_dir, EXPLORE_FILE_STRUCTURE.traj))
                kpu_list.append(kpu_dir)
        
        self.make_kpu_slurm_job_files(kpu_list)

    def make_kpu_slurm_job_files(self, kpu_list:list[str]):
        del_file_list_by_patten(self.kpu_dir, "*{}".format(TRAIN_FILE_STRUCTUR.kpu_job))
        group_list = split_job_for_group(self.resource.train_resource.group_size, kpu_list, 1)
        for group_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            jobname = "kpu{}".format(group_index)
            tag_name = "{}-{}".format(group_index, TRAIN_FILE_STRUCTUR.kpu_tag)
            tag = os.path.join(self.kpu_dir, tag_name)
            run_cmd = self.set_kpu_cmd()
            group_slurm_script = set_slurm_script_content(gpu_per_node=self.resource.train_resource.gpu_per_node, 
                number_node = self.resource.train_resource.number_node, 
                cpu_per_node = self.resource.train_resource.cpu_per_node,
                queue_name = self.resource.train_resource.queue_name,
                custom_flags = self.resource.train_resource.custom_flags,
                env_script = self.resource.train_resource.env_script,
                job_name = jobname,
                run_cmd_template = run_cmd,
                group = group,
                job_tag = tag,
                task_tag = TRAIN_FILE_STRUCTUR.kpu_tag, 
                task_tag_faild = TRAIN_FILE_STRUCTUR.kpu_tag_failed,
                parallel_num=1,
                check_type=None
                )
            slurm_script_name = "{}-{}".format(group_index, TRAIN_FILE_STRUCTUR.kpu_job)
            slurm_job_file = os.path.join(self.kpu_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    def set_kpu_cmd(self):
        script = ""
        pwmlff = self.resource.train_resource.command
        script += "{} {} -m {} -c {} -f {} -a {} -s {} >> {}\n\n".format(\
            pwmlff, MODEL_CMD.kpu, \
            TRAIN_FILE_STRUCTUR.dp_model_name, EXPLORE_FILE_STRUCTURE.traj, \
            LAMMPS.traj_format, LAMMPS.atom_type_file, EXPLORE_FILE_STRUCTURE.kpu_model_devi, SLURM_OUT.kpu_out)
        return script

        '''
    description: 
        waiting: if need set group size, make new script: work1 wait; work2 wait; ...
    param {*} self
    param {list} md_work_list
    return {*}
    author: wuxingxing
    '''    
    def do_kpu_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.kpu_dir, \
            job_patten="*-{}".format(TRAIN_FILE_STRUCTUR.kpu_job), \
            tag_patten="*-{}".format(TRAIN_FILE_STRUCTUR.kpu_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False
        if slurm_done is False:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these KPU Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), TRAIN_FILE_STRUCTUR.kpu_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished(error_type=SLURM_OUT.kpu_out)
                # mission.move_slurm_log_to_slurm_work_dir()
                    
    '''
    description: 
        iter.0000/explore/select: this is kpu work dir
        -------------------/md.000.sys.000/
        ----------------------------------/md.000.sys.000.t.000(p.000)
        -------------------------------------------------------------/kpu.job kpu.json dp_model.ckpt
        -------------------------------------------------------------kpu dir
        --------------------------------------------------------------------------/kpu_info.csv
    param {*} self
    return {*}
    author: wuxingxing
    '''
    def post_process_kpu(self):
        summary = select_image(
                md_dir=self.kpu_dir, 
                save_dir=self.select_dir,
                md_job=self.md_job,
                devi_name=EXPLORE_FILE_STRUCTURE.get_devi_name(UNCERTAINTY.kpu),
                lower=self.input_param.strategy.kpu_lower,  
                higer=self.input_param.strategy.kpu_upper
        )
        return summary




import os
from pwact.active_learning.user_input.scf_param import DFTInput
from pwact.active_learning.user_input.iter_input import SCFParam
from pwact.utils.file_operation import str_list_format
from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.utils.constant import DFT_STYLE, PWDATA
from pwact.utils.app_lib.pwmat import read_and_check_etot_input
class InitBulkParam(object):
    def __init__(self, json_dict: dict) -> None:
        self.root_dir = get_parameter("work_dir", json_dict, "./")
        if not os.path.isabs(self.root_dir):
            self.root_dir = os.path.realpath(self.root_dir)
        
        self.data_shuffle = get_parameter("data_shuffle", json_dict, False)
        self.data_format = get_parameter("data_format", json_dict, PWDATA.extxyz).lower()
        if self.data_format not in [PWDATA.extxyz, PWDATA.pwmlff_npy]:
            raise Exception("ERROR! The 'data_format' only support 'extxyz' or 'pwmlff/npy' format! The fefault format is 'extxyz'!")        
        self.interval = get_parameter("interval", json_dict, 1)

        # self.reserve_pwmat_files = get_parameter("reserve_pwmat_files", json_dict, False)
        self.reserve_work = get_parameter("reserve_work", json_dict, False)
        # read init configs
        sys_config_prefix = get_parameter("sys_config_prefix", json_dict, None)
        sys_configs = get_required_parameter("sys_configs", json_dict)
        if isinstance(sys_configs, dict):
            sys_configs = [sys_configs]

        # set sys_config detail
        self.dft_style = get_parameter("dft_style", json_dict, "PWMAT").lower()
        self.scf_style = get_parameter("scf_style", json_dict, None)

        self.sys_config:list[Stage] = []
        self.is_relax = False
        self.is_aimd = False
        self.is_scf = False
        self.is_bigmodel=False
        self.is_direct = False
        for index, config in enumerate(sys_configs):
            stage = Stage(config, index, sys_config_prefix, self.dft_style)
            self.sys_config.append(stage)
            if stage.relax:
                self.is_relax = True
            if stage.aimd:
                self.is_aimd = True
            if stage.scf:
                self.is_scf = True
            if stage.bigmodel:
                self.is_bigmodel = True
            if stage.direct:
                self.is_direct = True
                
        # for PWmat: set etot.input files and persudo files
        # for Vasp: set INCAR files and persudo files
        self.dft_input = SCFParam(json_dict=json_dict, 
                                    is_scf=self.is_scf, 
                                    is_relax=self.is_relax, 
                                    is_aimd=self.is_aimd, 
                                    root_dir=self.root_dir, 
                                    dft_style=self.dft_style, 
                                    scf_style=self.scf_style,
                                    is_bigmodel=self.is_bigmodel,
                                    is_direct=self.is_direct)
        
        # check and set relax etot.input file
        for config in self.sys_config:
            if self.is_relax:
                if config.relax_input_idx >= len(self.dft_input.relax_input_list):
                    raise Exception("Error! for config '{}' 'relax_input_idx' {} not in 'relax_input'!".format(os.path.basename(config.config_file), config.relax_input_idx))
                config.set_relax_input_file(self.dft_input.relax_input_list[config.relax_input_idx])
            
            if self.is_scf:
                if not os.path.exists(self.dft_input.scf_input_list[0].input_file):
                    raise Exception("Error! relabel dft input file {} not exisit!".format(self.dft_input.scf_input_list[0].input_file))
                config.set_scf_input_file(self.dft_input.scf_input_list[0])
            
            if self.is_bigmodel:
                if config.bigmodel_input_idx >= len(self.dft_input.bigmodel_input_list):
                    raise Exception("Error! for script '{}' 'bigmodel_input_idx' {} not in 'bigmodel_input'!".format(os.path.basename(config.config_file), config.bigmodel_input_idx))
                config.set_bigmodel_input_file(self.dft_input.bigmodel_input_list[config.bigmodel_input_idx])

            if self.is_direct:
                if config.direct_input_idx >= len(self.dft_input.direct_input_list):
                    raise Exception("Error! for script '{}' 'direct_input_idx' {} not in 'direct_input'!".format(os.path.basename(config.config_file), config.direct_input_idx))
                config.set_direct_input_file(self.dft_input.direct_input_list[config.direct_input_idx])

            # check and set aimd etot.input file
            if self.is_aimd:
                if config.aimd_input_idx >= len(self.dft_input.aimd_input_list):
                    raise Exception("Error! for config '{}' 'aimd_input_idx' {} not in 'aimd_input'!".format(os.path.basename(config.config_file), config.aimd_input_idx))
                config.set_aimd_input_file(self.dft_input.aimd_input_list[config.aimd_input_idx])

class Stage(object):
    def __init__(self, json_dict: dict, index:int, sys_config_prefix:str = None, dft_style:str=None) -> None:
        self.dft_style = dft_style #not used
        self.config_index = index
        self.use_dftb = False
        self.use_skf = False
        
        config_file = get_required_parameter("config", json_dict)
        self.config_file = os.path.join(sys_config_prefix, config_file) if sys_config_prefix is not None else config_file
        if not os.path.exists:
            raise Exception("ERROR! The sys_config {} file does not exist!".format(self.config_file))
        self.format = get_parameter("format", json_dict, PWDATA.pwmat_config).lower()
        self.pbc = get_parameter("pbc", json_dict, [1,1,1])
        # extract config file to Config object, then use it
        self.relax = get_parameter("relax", json_dict, False)
        self.relax_input_idx = get_parameter("relax_input_idx", json_dict, 0)
        self.relax_input_file = None
        
        self.aimd = get_parameter("aimd", json_dict, False)
        self.aimd_input_idx = get_parameter("aimd_input_idx", json_dict, 0)
        self.aimd_input_file = None
        
        self.scf = get_parameter("scf", json_dict, False)
        self.scf_input_idx = get_parameter("scf_input_idx", json_dict, 0)
        self.scf_input_file = None

        self.bigmodel = get_parameter("bigmodel", json_dict, False)
        self.bigmodel_input_idx = get_parameter("bigmodel_input_idx", json_dict, 0)
        self.bigmodel_script = None

        self.direct = get_parameter("direct", json_dict, False)
        self.direct_input_idx = get_parameter("direct_input_idx", json_dict, 0)
        self.direct_script = None
        
        if self.bigmodel and self.aimd:
            raise Exception("ERROR! The 'aimd' and 'bigmodel' cannot be set simultaneously!")
        
        super_cell = get_parameter("super_cell", json_dict, [])
        super_cell = str_list_format(super_cell)
        if len(super_cell) > 0:
            if len(super_cell) == 3 :#should c
                if isinstance(super_cell[0], int):
                    super_cell = [[super_cell[0], 0, 0], [0, super_cell[1], 0], [0, 0, super_cell[2]]]
            elif len(super_cell) == 9:
                super_cell = [super_cell[0:3], super_cell[3:6], super_cell[6:9]]
            else:
                raise Exception("Error! The input super_cell should be 3 or 9 values, for example:\
                    list format [1,2,1], [1,0,0,0,2,0,0,0,3]!")
            self.super_cell = super_cell
        else:
            self.super_cell = None
        
        scale = get_parameter("scale", json_dict, [])
        scale = str_list_format(scale)
        if len(scale) > 0:
            self.scale = scale
        else:
            self.scale = None
            
        self.perturb = get_parameter("perturb", json_dict, 0)
        if self.perturb == 0:
            self.perturb = None
        self.cell_pert_fraction = get_parameter("cell_pert_fraction", json_dict, 0.03)
        self.atom_pert_distance = get_parameter("atom_pert_distance", json_dict, 0.01)
    
    def set_relax_input_file(self, input_file:DFTInput):
        self.relax_input_file = input_file.input_file
        self.relax_kspacing = input_file.kspacing 
        self.relax_flag_symm = input_file.flag_symm

    def set_scf_input_file(self, input_file:DFTInput):
        self.scf_input_file = input_file.input_file
        self.scf_kspacing = input_file.kspacing 
        self.scf_flag_symm = input_file.flag_symm

    def set_aimd_input_file(self, input_file:DFTInput):
        self.aimd_input_file = input_file.input_file
        self.aimd_kspacing = input_file.kspacing
        self.aimd_flag_symm = input_file.flag_symm
        self.use_dftb = input_file.use_dftb
        self.use_skf = input_file.use_skf

    def set_bigmodel_input_file(self, input_file:DFTInput):
        self.bigmodel_input_file = input_file.input_file
        self.bigmodel_kspacing = input_file.kspacing 
        self.bigmodel_flag_symm = input_file.flag_symm

    def set_direct_input_file(self, input_file:DFTInput):
        self.direct_input_file = input_file.input_file
        self.direct_kspacing = input_file.kspacing 
        self.direct_flag_symm = input_file.flag_symm
